# coding: utf-8
"""
    VoxelUtils.py

    from https://www.mahirokazuko.com/entry/2019/06/24/232831
"""
import colorsys
import collections
import numpy as np
from matplotlib import pyplot as plt
from mpl_toolkits.mplot3d import Axes3D

def int2hue(integer, in_max=255, out_max=255, is_alpha=False):
    if is_alpha:
        return tuple( float(v * out_max) for v in  colorsys.hsv_to_rgb(integer/in_max, 1, 1) ) + (out_max,)
    else:
        return tuple( float(v * out_max) for v in  colorsys.hsv_to_rgb(integer/in_max, 1, 1) )

def flatten(xs):
    result = []
    for x in xs:
        if isinstance(x, collections.Iterable) and not isinstance(x, dict)and not isinstance(x, str):
            result.extend(flatten(x))
        else:
            result.append(x)
    return result

def draw_voxel(ax, tensor):
    tensor = tensor.transpose(2,0,1) # 軸を入れ替える。※ 機械学習で画像を扱うときによく使う軸の順番にしただけで、それ以外の深い意味は無い
    # fig = plt.figure(figsize=(5, 5), dpi=100) # 図の大きさを変えられる
    # ax = fig.gca(projection='3d')
    org_shape = tensor.shape

    tensor_min = tensor.min() # すべての要素が正になるように、もし負の値があったらその絶対値を全要素に加算する
    if tensor_min < 0:
        tensor = tensor - tensor_min

    eps = 1e-10
    tensor = tensor + eps # 要素の値が0だと表示できないことがあるので、それを回避するために微小な数を与える

    tensor_max = tensor.max()
    colors = [int2hue(x, in_max=tensor_max, out_max=1.0, is_alpha=True) for x in flatten(tensor)]
    colors = np.reshape(colors, org_shape + (4,)) # 各座標に、RGBAを表すサイズ4の配列を埋め込むためシェイプに4を追加する

    if False:
        # 立方体になるように軸を調整
        xsize, ysize, zsize = tensor.shape
        max_range = np.array([xsize, ysize, zsize]).max() * 0.5
        mid_x = xsize * 0.5
        mid_y = ysize * 0.5
        mid_z = zsize * 0.5
        ax.set_xlim(mid_x - max_range, mid_x + max_range)
        ax.set_ylim(mid_y - max_range, mid_y + max_range)
        ax.set_zlim(mid_z - max_range, mid_z + max_range)

        # 目盛りを削除
        plt.setp(ax.get_xticklabels(), visible=False)
        plt.setp(ax.get_yticklabels(), visible=False)
        plt.setp(ax.get_zticklabels(), visible=False)

    ax.voxels(tensor, facecolors=colors, linewidth=0.3, edgecolor='k')
    #  plt.show()



"""
    BaseSurfaceSpline.py

    Copyright (c) 2019, SAXS Team, KEK-PF
"""
import numpy as np
from scipy.optimize import minimize
from mpl_toolkits.mplot3d import Axes3D
from molass_legacy.KekLib.ThreeDimUtils import compute_plane
from .AveragingDiscussion import extrapolate
import molass_legacy.KekLib.DebugPlot as plt

def compute_concentration(a, b, c, sign=1):
    return ( -b + sign*np.sqrt(b**2 - 4*a*c) ) / (2*a)

class BaseSurfaceSpline:
    def __init__(self, sd):
        self.q = sd.intensity_array[0,:,0]
        self.data = sd.intensity_array[:,:,1].T
        self.e_curve = sd.xray_curve

        """
        argmin(a,b,c,d,e,f,g) ||AB・W - M_||
        such that

        AB = [d*x+e, f*x+g].T

        z = a*x + b*y + c
        BP = compute_plane(a, b, c, x, y)

        M_ = M - BP
        M_(x,j) = (d*x+e)*u[j] + (f*x+g)*u[j]**2
        u = compute_concentration(f*x+g, d*x+e, -M_(x,:))

        U = [u, u**2]

        """
        peak_i = self.e_curve.primary_peak_i
        xslice = slice(0,100)
        yslice = slice(peak_i-50, peak_i+50)
        y = np.arange(self.data.shape[1])[yslice]

        u   = self.e_curve.y[yslice]
        u_  = u/np.max(u)

        wsize = 10
        w_list = []
        abc_list = []
        for k in range(0, 100, wsize):
            wslice = slice(k,k+wsize)
            M   = self.data[wslice,yslice]
            P   = extrapolate(M, u_)
            Aq  = P[:,0]
            Bq  = P[:,1]
            d_init = np.average(Aq)
            e_init = np.average(Bq)
            init_params = np.array([0, 0, 0, d_init, e_init])

            w   = self.q[wslice]
            i   = wsize//2
            ones = np.ones(M.shape[0])
            def obj_func(p):
                BP = compute_plane(*p[0:3], w, y)
                M_ = M - BP
                r = compute_concentration(p[4], p[3], -M_[i,:])
                # r_ = r/np.max(r)
                r_ = r
                if False:
                    fig = plt.figure()
                    ax = fig.gca()
                    ax.plot(M_[i,:])
                    ax.plot(r_*np.max(M_[i,:]))
                    plt.show()
                R = np.array([r_, r_**2])
                AB = np.array([ones*p[3], ones*p[4]]).T
                ret = np.sum((AB @ R - M_).flatten()**2)
                # print('ret=', ret)
                return ret

            res = minimize(obj_func, init_params)
            a,b,c = res.x[0:3]
            w_list.append(w)
            abc_list.append([a,b,c])

        if True:
            x = self.q[xslice]
            xx, yy = np.meshgrid( x, y )
            i = np.arange(xslice.start, xslice.stop)
            j = np.arange(yslice.start, yslice.stop)
            ii, jj = np.meshgrid( i, j )

            zz = self.data[ii, jj]

            fig = plt.figure(figsize=(10, 8))
            ax = fig.add_subplot(111, projection='3d')
            ax.plot_surface(xx, yy, zz, color='green', alpha=0.3 )

            if False:
                x_ = self.q[wslice]
                y_ = np.ones(len(x_))*peak_i
                ax.plot(x_, y_, Aq, color='red')
                ax.plot(x_, y_, Bq, color='blue')

            for k, w in enumerate(w_list):
                a,b,c = abc_list[k]
                print([k], a, b, c)
                ww, yy = np.meshgrid( w[[0,-1]], y[[0,-1]] )
                zz = a*ww + b*yy + c
                ax.plot_surface(ww, yy, zz, color='red', alpha=0.3 )

            fig.tight_layout()
            plt.show()



# coding: utf-8
"""
    GammaVisualizer.py

    Copyright (c) 2020, SAXS Team, KEK-PF
"""
import numpy as np
from mpl_toolkits.mplot3d import Axes3D
from matplotlib.collections import PolyCollection
# from matplotlib.patches import Polygon
from matplotlib import colors as mcolors
import molass_legacy.KekLib.DebugPlot as plt
from molass_legacy.KekLib.OurMatplotlib import get_facecolors

class GammaVisualizer:
    def __init__(self):
        pass

    def draw_gamma_3d(self, ax, gamma):
        num_components = gamma.shape[1]
        zs = np.arange(num_components)
        xs = np.arange(gamma.shape[0])
        verts = []
        for k in range(num_components):
            ys = gamma[:,k]
            # ys[0], ys[-1] = 0, 0
            verts.append([(0,0)] + list(zip(xs, ys)) + [(xs[-1], 0)])
        poly = PolyCollection(verts, facecolors=get_facecolors(num_components, 2))
        poly.set_alpha(0.7)
        ax.add_collection3d(poly, zs=zs, zdir='y')
        ax.set_xlim3d(0, xs[-1])
        ax.set_ylim3d(-0.5, num_components-0.5)
        ax.set_zlim3d(0, 1)
        return poly, verts

    def show(self, gamma):
        plt.push()
        fig = plt.figure(figsize=(18,8))
        ax = fig.add_subplot(111, projection='3d')
        ax.set_title(r'$\gamma$ Visualization', fontsize=20)
        self.draw_gamma_3d(ax, gamma)
        fig.tight_layout()
        plt.show()
        plt.pop()



"""
    Wave.ScatteringN.py

    Copyright (c) 2020, SAXS Team, KEK-PF
"""
import numpy as np
import matplotlib.pyplot as plt
import matplotlib.cm as cm
import matplotlib.animation as animation
from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg
from molass_legacy.KekLib.OurTkinter import Tk, Dialog
from molass_legacy.KekLib.OurMatplotlib import NavigationToolbar

class BeamAnimDialog(Dialog):
    def __init__(self, parent):
        Dialog.__init__(self, parent, "Beam Animatio", visible=False)

    def show(self):
        self._show()

    def body(self, body_frame):
        cframe = Tk.Frame(body_frame)
        cframe.pack()
        bframe = Tk.Frame(body_frame)
        bframe.pack(fill=Tk.BOTH, expand=1)
        tframe = Tk.Frame(bframe)
        tframe.pack(side=Tk.LEFT)
        pframe = Tk.Frame(bframe)
        pframe.pack(side=Tk.RIGHT)

        fig, axes = plt.subplots(ncols=2, figsize=(14, 7))
        self.fig  = fig
        self.mpl_canvas = FigureCanvasTkAgg(fig, cframe)
        self.mpl_canvas_widget = self.mpl_canvas.get_tk_widget()
        self.mpl_canvas_widget.pack(fill=Tk.BOTH, expand=1)
        self.toolbar = NavigationToolbar(self.mpl_canvas, tframe)
        self.toolbar.update()
        self.draw_anim(fig, axes)
        self.mpl_canvas.draw()

    def draw_anim(self, fig, axes):
        ax1, ax2 = axes

        for ax in axes:
            ax.grid(False)

        N = 800
        n = N//2
        r = np.arange(N)
        rr = r**2
        i = np.arange(N)
        ii, jj = np.meshgrid(i, i)

        d = 15

        target_points = [(n,n), (n-d,n-d), (n+d,n+d), (n-d,n+2*d)]

        labels_list = []
        target_particles = []
        for point in target_points:
            p, q = point
            labels = np.searchsorted(rr, (ii - p)**2 + (jj-q)**2)
            labels_list.append(labels)
            target_particles.append(labels <= 5)

        canvas = np.zeros((N,N))

        def color_filter(v, maxv):
            return 1/(1 + np.exp(-2*(v - (maxv-1))))

        cmap = 'plasma'

        Lmax = np.max(labels_list)
        x = np.linspace(0, np.pi*96, N+Lmax)

        num_points = len(target_points)

        def animate_func(i):
            t = np.pi*i/2

            intensity = np.zeros(canvas.shape)
            for point, labels, particle in zip(target_points, labels_list, target_particles):
                p, q = point

                beam = color_filter(1 + np.sin(x[0:p] - t), 2)
                canvas[q-2:q+3, 0:p] = beam  # as if linewith=5
                canvas[particle] = 1

                scattered = 1 + np.sin(x[p:] - t)
                intensity += scattered[labels]

            im1 = ax1.imshow(canvas, animated=True, cmap=cmap, origin='lower')
            im2 = ax2.imshow(color_filter(intensity, num_points*2), animated=True, cmap=cmap, origin='lower')
            return [im1, im2]

        fig.tight_layout()
        self.ani = animation.FuncAnimation(fig, animate_func, frames=60, interval=50, blit=True)
        # self.ani.save('scattering.mp4', writer="ffmpeg")

def demoN():
    from molass_legacy.KekLib.TkUtils import get_tk_root

    root = get_tk_root()

    def show_demo():
        dialog = BeamAnimDialog(root)
        dialog.show()
        root.quit()

    root.after(100, show_demo)
    root.mainloop()
    root.destroy()



"""
    RangeSpecFrame.py

    Copyright (c) 2019-2024, SAXS Team, KEK-PF
"""
import os
import copy
import numpy                as np
import matplotlib.pyplot    as plt
from matplotlib.gridspec    import GridSpec
from matplotlib import colors
import matplotlib
import seaborn as sns
from matplotlib.backends.backend_tkagg  import FigureCanvasTkAgg
from molass_legacy.KekLib.OurTkinter             import Tk
from molass_legacy.KekLib.TkUtils                import split_geometry
from molass_legacy.KekLib.OurMatplotlib          import NavigationToolbar, get_color, get_hex_color
from molass_legacy.KekLib.DisguinsingWidgets     import CheckonLabel
from molass_legacy._MOLASS.SerialSettings         import get_setting, set_setting
from molass_legacy.DataStructure.PeakInfo               import PeakInfo
from molass_legacy.DataStructure.AnalysisRangeInfo      import PairedRange

RANGE_FRAME_HEIGHT      = 60

class RangeSpecFrame(Tk.LabelFrame):
    def __init__(self, parent, *args, **kwargs):
        self.parent = parent

        self.editor = kwargs.pop('editor', None)
        self.j_min = kwargs.pop('j_min', 0)
        self.j_max = kwargs.pop('j_max', 1000)
        self.peak_no = kwargs.pop('peak_no', None)
        self.active_peak_no = kwargs.pop('active_peak_no', None)
        self.row_no_base =  kwargs.pop('row_no_base', None)
        self.col_widths = col_widths = kwargs.pop('col_widths', None)
        self.select_list = kwargs.pop('select_list', None)
        self.range_list = kwargs.pop('range_list', None)
        """
            self.range_list == self.editor.editor_ranges[k]
        """
        self.peak = kwargs.pop('peak', None)
        self.peak_info = PeakInfo(self.peak_no-1, self.peak.top_x)

        num_ranges_ = len(self.range_list)

        Tk.LabelFrame.__init__(self, parent, *args, **kwargs)

        # Peak No
        frame1 = Tk.Frame(self, width=col_widths[1], height=20)
        frame1.pack_propagate(0)
        frame1.pack(side=Tk.LEFT)
        label = Tk.Label(frame1, text=str(self.peak_no))
        label.pack()

        # Number of Ranges
        frame3 = Tk.Frame(self, width=col_widths[2], height=20)
        frame3.pack_propagate(0)
        frame3.pack(side=Tk.LEFT)
        self.num_ranges = Tk.IntVar()
        self.num_ranges.set(num_ranges_)
        min_num_ranges = 1
        # state = Tk.NORMAL if self.peak.sign > 0 else Tk.DISABLED
        state = Tk.NORMAL
        self.spinbox = Tk.Spinbox( frame3, textvariable=self.num_ranges,
                                from_=min_num_ranges, to=2, increment=1,
                                justify=Tk.CENTER, width=6, state=state )

        self.num_ranges.trace('w', self.num_ranges_tracer)
        self.spinbox.pack()

        # Range(s)
        self.frame34 = frame34 = Tk.Frame(self, width=col_widths[3]+col_widths[4], height=RANGE_FRAME_HEIGHT)
        frame34.pack_propagate(0)
        frame34.pack(side=Tk.LEFT)

        # Ignore
        frame6 = Tk.Frame(self, width=col_widths[5], height=20)
        frame6.pack_propagate(0)
        frame6.pack(side=Tk.LEFT)
        self.ignore = Tk.IntVar()
        self.ignore.set(int(self.editor.ignorable_flags[self.peak_no-1]))

        self.range_widgets = []
        self.update_ranges()    # uses self.ignore

        if self.peak.sign > 0:
            cb = Tk.Checkbutton(frame6, variable=self.ignore)
            self.ignore.trace('w', self.ignore_tracer)
        else:
            cb = CheckonLabel(frame6)
        cb.pack()

    def update_ranges(self, *args):
        for w in self.range_widgets:
            w.destroy()

        peak = self.peak
        num_ranges = self.num_ranges.get()
        range_list = self.range_list

        self.range_widgets = []
        """
            gridding directly from self.frame34 seems to break self.frame34.pack_propagate(0)
            therefore, create an intermediate frame and do gridding from it
        """
        frame = Tk.Frame(self.frame34)
        frame.pack(fill=Tk.BOTH, expand=1)
        self.range_widgets.append(frame)

        self.range_var_list = []
        self.edit_buttons = []
        for k in range(num_ranges):
            frame.grid_rowconfigure(k, weight=1)
            range_frame = Tk.Frame(frame)
            range_frame.grid(row=k, column=0)

            label_frame = Tk.Frame(range_frame, width=self.col_widths[3], height=20)
            label_frame.pack_propagate(0)
            label_frame.pack(side=Tk.LEFT)
            text = '' if self.active_peak_no is None else ('%d-%d' % (self.active_peak_no, k+1))
            label = Tk.Label(label_frame, text=text)
            label.pack()

            spinbox_frame = Tk.Frame(range_frame, width=self.col_widths[4], height=30)
            spinbox_frame.pack_propagate(0)
            spinbox_frame.pack(side=Tk.LEFT, padx=5)

            fv, tv = range_list[k]

            f = Tk.IntVar()
            f.set(fv)
            t = Tk.IntVar()
            t.set(tv)
            self.range_var_list.append( [f, t] )
            spinbox1 = Tk.Spinbox( spinbox_frame, textvariable=f,
                                from_=self.j_min, to=self.j_max, increment=1,
                                justify=Tk.CENTER, width=6 )
            spinbox1.pack(side=Tk.LEFT, padx=5, pady=5)
            f.trace( 'w',  lambda *args, k_=k: self.spinbox_tracer(k_, 0) )

            spinbox2 = Tk.Spinbox( spinbox_frame, textvariable=t,
                                from_=self.j_min, to=self.j_max, increment=1,
                                justify=Tk.CENTER, width=6 )
            spinbox2.pack(side=Tk.LEFT, padx=5, pady=5)
            t.trace( 'w',  lambda *args, k_=k: self.spinbox_tracer(k_, 1) )
            self.spinbox_trace = True

            state = Tk.DISABLED if self.ignore.get() else Tk.NORMAL
            edit_button = Tk.Button(spinbox_frame, text="🖉", state=state,
                                command=lambda k_=k: self.show_range_inspector(k_))
            edit_button.pack(side=Tk.LEFT, padx=5)
            self.edit_buttons.append(edit_button)

    def spinbox_tracer(self, i, j):
        if not self.spinbox_trace:
            return

        try:
            v = self.range_var_list[i][j].get()
            # print( 'spinbox_tracer: ', (i, j), v )
            self.range_list[i][j] = v
            if False:
                self.editor.update_figs(peak_no=self.peak_no)
            else:
                self.editor.refresh_figs()
        except:
            pass

    def get_conc_depend(self):
        return self.editor.dialog.get_conc_depend()

    def show_range_inspector(self, k):
        from molass_legacy.DataStructure.AnalysisRangeInfo import PairedRange
        from molass_legacy.PeaksetSelector import PeakSetSelector
        from RangeInspector import RangeInspectorDialog
        row = self.row_no_base + k
        print('show_range_inspector: (row, k)=', (row, k))
        print('select_list=', self.select_list)
        print('range_list=', self.range_list)
        solver = self.editor.get_extrapolation_solver()
        j0 = self.editor.xr_j0
        ranges = [[j-j0 for j in fromto] for fromto in self.range_list]
        paired_range = PairedRange(self.peak_info, *ranges)
        selector = PeakSetSelector(solver.cnv_ranges, self.editor.mapper.x_curve)
        peakset_info = selector.select_peakset(row)
        ad = k
        conc_depend = self.get_conc_depend()
        inspector = RangeInspectorDialog(self.editor.dialog.parent, solver, j0,
                        paired_range, peakset_info, ad, conc_depend=conc_depend)
        ret = inspector.show()
        if ret:
            new_shifted_range = inspector.get_range(shifted=True)
            self.editor.update_range(self.peak_no-1, ad, *new_shifted_range)

    def update_range(self, ad, f, t):
        self.spinbox_trace = False
        vars_ = self.range_var_list[ad]
        vars_[0].set(f)
        vars_[1].set(t)
        self.update()
        self.spinbox_trace = True

    def num_ranges_tracer(self, *args):
        try:
            num_ranges = self.num_ranges.get()
        except:
            return

        assert num_ranges in [0, 1, 2]

        prev_range_list = copy.deepcopy(self.range_list)
        self.range_list.clear()
        """
            the above also means self.editor.editor_ranges[k].clear()
            because self.range_list == self.editor.editor_ranges[k]
        """

        if num_ranges > 0:
            if len(prev_range_list) > 0:
                f = prev_range_list[0][0]
                t = prev_range_list[-1][1]
            else:
                range_list = self.opt_rec.get_range_list(self.editor.x)
                f = range_list[0][0]
                t = range_list[-1][-1]

            if num_ranges == 1:
                self.range_list.append( [f, t] )
            elif num_ranges == 2:
                top_x = self.j_min + self.peak.top_x
                self.range_list.append( [f, top_x] )
                self.range_list.append( [top_x, t] )

        self.update_ranges()

        # specifying one peak_no is not appropriate, it is better to update all figgures
        self.editor.refresh_figs()
        self.editor.refresh_spec_panels()

    def update_eebars(self):
        for panel in self.editor.specpanel_list:
            panel.eebar.refresh()

    def ignore_tracer(self, *args):
        ignore = bool(self.ignore.get())
        self.editor.ignorable_flags[self.peak_no-1] = ignore
        state = Tk.DISABLED if ignore else Tk.NORMAL
        for w in self.edit_buttons:
            w.config(state=state)
        self.editor.refresh_figs()
        self.editor.refresh_spec_panels()

    def get_info_for_reset(self):
        num_ranges = self.num_ranges.get()
        range_info = [ (v[0].get(), v[1].get()) for v in self.range_var_list]
        ignore = self.ignore.get()
        return (num_ranges, range_info, ignore)

    def get_paired_range(self):
        # note that elm_recs is omitted here
        # consider to refactor to unify PeakInfo and EmgPeak
        range_list = [ [f - self.j_min, t - self.j_min] for f, t  in self.range_list]
        return PairedRange(self.peak_info, *range_list)



"""
module for EM algorithm in Gaussian mixture model
"""
import numpy as np
from model_gm import ModelGM

class EM():
    """
    class for EM algorithm
    """
    def __init__(self, k, sigma=None, tol=1e-3, max_iter=100, print_iter=False):
        """
        Args:
        tol: float
        stop criterion

        k: int
        number of components

        sigma: common standard deviation
        if sigma is None, it will be estimated

        tol: float
        stop criterion, tolerance of log-likelihood increase

        max_iter: int
        maximum number of interations

        print_iter: bool
        if True, will print more information on each iteration
        """
        self.k = k
        self.sigma = sigma
        self.tol = tol
        self.max_iter = max_iter
        self.print_iter = print_iter

    def estimate(self, samples, num_rd=5, x_range=None, s_range=None):
        """
        estimate a model
        best of num_rd random initial guesses
        initial means are uniform from x_range
        initial sigma are uniform from s_range

        Args:
        num_rd(int): number of random initial guess, default 5
        x_range [a,b]: initial guess range of means, default [-1,1]
        s_range [c,d]: initial guess range of sigmas, default [0.5,1.5]
        """
        if x_range is None:
            x_range = [-1, 1]
        if s_range is None:
            s_range = [0.5, 1.5]

        ll_max = float('-inf')
        for _ in range(num_rd):
            w_init = np.random.dirichlet(np.ones(self.k))
            x_init = np.random.uniform(x_range[0], x_range[1], self.k)
            if self.sigma is None:
                s_init = np.random.uniform(s_range[0], s_range[1])
            else:
                s_init = self.sigma
            start = ModelGM(w=w_init, x=x_init, std=s_init)
            model_cur, _, ll_cur = self.estimate_with_init(samples, start, detail=True)
            if ll_cur > ll_max:
                model = model_cur
                ll_max = ll_cur
        return model

    def estimate_with_init(self, samples, init, detail=False):
        """
        estimate a model from a given initial
        Args:
        init (modelGM): initial guess

        Returns:
        model(modelGM): estimated model
        iterN(int): number of iterations
        ll_cur(float): last log-likelihood
        """
        # assert self.k == len(init.weights)
        k_cur = len(init.weights)
        num = len(samples)
        samples = np.asarray(samples)

        num_iter = 0
        model = init
        l_mat = np.exp(ll_mat(samples, model)) # shape (n,k)
        ll_cur = np.sum(np.log(np.dot(l_mat, model.weights)))

        while True:
            ll_pre = ll_cur

            labels = l_mat * model.weights # shape (n,k)
            labels /= np.sum(labels, axis=1)[:, np.newaxis]

            sum_labels = np.sum(labels, axis=0) # shape (k,)

            num_iter += 1
            model.weights = sum_labels/num
            model.centers = np.dot(samples, labels)/sum_labels

            if self.sigma is None:
                # EM iteration of estimating the common variance
                cross = model.centers**2-2*np.outer(samples, model.centers)\
                        +(samples**2)[:, np.newaxis]
                sigma2 = np.sum(cross*labels)/num
                model.sigma = np.ones(k_cur) * np.sqrt(sigma2)

            l_mat = np.exp(ll_mat(samples, model))
            ll_cur = np.sum(np.log(np.dot(l_mat, model.weights)))

            if self.print_iter:
                print(model.weights, model.centers, model.sigma)
                print(ll_cur)

            if num_iter > self.max_iter or ll_cur-ll_pre < self.tol:
                break

        if detail:
            return model, num_iter, ll_cur
        else:
            return model

def ll_mat(samples, model):
    """
    log-likelihood of samples

    Args:
    samples: ndarray of length n
    model: ModelGM instance of k components with common sigma

    Returns:
    matrix of log-likelihoods of shape (n,k)
    """
    samples = np.asarray(samples)
    precision = 1./(model.sigma**2) # inverse of variance (sigma^2)
    ll0 = model.centers**2*precision - 2*np.outer(samples, model.centers*precision) \
          +np.outer(samples**2, precision)
    return -0.5*(np.log(2*np.pi)+ll0)-np.log(model.sigma)


def ll_sample(samples, model):
    """
    Log-likelihood matrix of samples under the given GM model

    Args:
    samples: ndarray of length n
    model: ModelGM instance of k components with common sigma

    Return:
    log-likelihood of all samples
    """
    return np.sum(np.log(np.dot(np.exp(ll_mat(samples, model)), model.weights)))



"""
    DecompEditorDialog.py

    Copyright (c) 2018-2025, SAXS Team, KEK-PF
"""
import os
import logging
from molass_legacy.KekLib.OurTkinter import Tk, Dialog, ttk
from molass_legacy.KekLib.TkSupplements import set_icon
from molass_legacy.KekLib.TkUtils import is_low_resolution
from molass_legacy.Decomposer.DecompUtils import CorrectedBaseline
from .DecompDummyFrame import DecompDummyFrame
from molass_legacy._MOLASS.SerialSettings import get_setting, set_setting
from molass_legacy._MOLASS.Version import is_developing_version
from molass_legacy._MOLASS.Version import is_developing_version
from molass_legacy.Selective.ModelSelectFrame import enable_edm_model, MODEL_NAMES

class DecompEditorDialog(Dialog):
    def __init__(self, parent, title, sd, mapper, judge_holder, scrollable=False, debug=False):
        self.logger  = logging.getLogger( __name__ )

        if debug:
            from molass_legacy.Mapping.MappingUtils import save_mapped_curves
            save_mapped_curves(mapper, plot=True)

        self.parent = parent
        self.title_ = title
        self.sd     = sd
        self.serial_data = sd   # for compatibility with AnalyzerDialog for Global Fitting
        self.mapper = mapper
        self.judge_holder = judge_holder
        self.ready = False
        self.scrollable = scrollable or is_low_resolution()
        self.canceled = False
        self.applied = False
        self.init_use_elution_models = get_setting('use_elution_models')
        set_setting('use_elution_models', 1)    # required to preview in this dialog
        self.global_fit = Tk.IntVar()
        self.global_fit.set(0)
        self.global_fit.trace_add("write", self.global_fit_tracer)
        self.global_frames = [None, None]

    def get_model_list(self):
        if is_developing_version():
            from importlib import reload
            import molass_legacy.Models.ElutionCurveModels
            import molass_legacy.Models.RateTheory.EDM
            # import molass_legacy.Models.Stochastic.Tripore
            reload(molass_legacy.Models.ElutionCurveModels)
            reload(molass_legacy.Models.RateTheory.EDM)
            # reload(molass_legacy.Models.Stochastic.Tripore)
        from molass_legacy.Models.ElutionCurveModels import EGHA, EMGA
        if enable_edm_model:
            from molass_legacy.Models.RateTheory.EDM import EDM
            from molass_legacy.Models.Stochastic.Monopore import Monopore
            MODEL_LIST = [EMGA(), EGHA(), EDM(delayed=True), Monopore(delayed=True)]
        else:
            MODEL_LIST = [EMGA(), EGHA()]
        return MODEL_LIST 

    def cancel(self):
        # overiding cancel to cleanup self.fig
        # because the call to the destructor __del__ seems to be delayed
        for frame in self.frames:
            frame.close_figs()
        print("DecompEditorDialog: closed figs")
        Dialog.cancel(self)

    def user_cancel(self):
        # this method is provided aside from "cancel" to avoid confusion.
        set_setting('use_elution_models', self.init_use_elution_models)
        self.cancel()

    def is_ready( self ):
        return self.ready

    def prepare( self, counter=None):
        self.parent.config( cursor='wait' )
        self.counter = counter
        self.parent.update()
        Dialog.__init__( self, self.parent, self.title_, visible=False, block=False )
        self.parent.config( cursor='' )
        self.ready = True
        self.advanced_frame.enable_buttons()

    def debug_sd(self):
        from molass_legacy.DataStructure.MdViewer import MdViewer
        sd = self.sd
        M = sd.intensity_array[:,:,1].T
        mdv = MdViewer(self, M, xvector=sd.qvector)
        mdv.show()

    def show(self, counter=None):
        if not self.ready:
            # for backward compatibility
            self.prepare(counter=counter)

        # self.after(1000, self.debug_sd)
        self._show()

    def body(self, default_body_frame, devel=True):
        if devel:
            from importlib import reload
            import molass_legacy.RangeEditors.DecompEditorFrame
            reload(molass_legacy.RangeEditors.DecompEditorFrame)
        from .DecompEditorFrame import DecompEditorFrame
        set_icon( self )

        if self.scrollable:
            from molass_legacy.KekLib.ScrolledFrame import ScrolledFrame
            self.scrolled_frame = ScrolledFrame(default_body_frame)
            # scrolled_frame.pack( fill=Tk.BOTH, expand=1 )
            self.scrolled_frame.pack(anchor=Tk.N)
            body_frame = self.scrolled_frame.interior
        else:
            self.scrolled_frame = None
            body_frame = default_body_frame

        self.body_frame = body_frame    # for use in create_new_frame

        self.corbase_info = CorrectedBaseline(self.sd, self.mapper)
        if self.counter is not None:
            self.counter[0] += 1
            self.parent.update()

        model_list = self.get_model_list()
        memorized_model = get_setting('editor_model')
        if memorized_model is None:
            selected_index = None
        else:
            selected_index = MODEL_NAMES.index(memorized_model)
            selected_model = model_list[selected_index]
            self.logger.info('previously applied %s has been selected.' % selected_model.get_name())

        min_score = None
        the_error = None
        min_index = None
        min_model = None
        self.frames = []
        scores = []
        errors = []
        for k, model in enumerate(model_list):
            if model.is_delayed() and (selected_index is None or selected_index != k):
                frame = DecompDummyFrame(body_frame, model)
            else:
                self.min_index = min_index      # for reference in delayed (i.e., non-traditional,) frames
                frame = DecompEditorFrame(body_frame, self, self.sd, self.mapper, self.corbase_info, model)
                if model.is_delayed():
                    model.set_delayed_off()
            frame.pack()
            if self.counter is not None:
                self.counter[0] += 1
                self.parent.update()
            self.frames.append(frame)
            score = frame.get_decomp_score()
            scores.append(score)
            error = frame.get_fit_error()
            errors.append(error)
            # print([k], error)
            if min_score is None or score < min_score:
                min_score = score
                the_error = error
                min_index = k
                min_model = model

        if selected_index is None:
            selected_index = min_index

        self.current = selected_index
        self.min_index = min_index      # for reference in delayed frames
        for k in range(len(model_list)):
            if k != self.current:
                self.frames[k].pack_forget()

    def buttonbox(self, devel=False):
        from molass_legacy.Selective.ModelSelectFrame import ModelSelectFrame
        from molass_legacy.Selective.AdvancedFrame import AdvancedFrame
        from molass_legacy.Extrapolation.PreviewButtonFrame import PreviewButtonFrame

        box = Tk.Frame(self)
        box.pack(fill=Tk.X)

        lower_frame = Tk.Frame(box)
        lower_frame.pack(fill=Tk.X)

        additional_box = Tk.Frame(lower_frame)
        additional_box.pack(side=Tk.RIGHT, anchor=Tk.N + Tk.E)
        standard_box = Tk.Frame(lower_frame)
        standard_box.pack(anchor=Tk.CENTER, padx=10)

        self.guide_frame = Tk.Frame(box)
        self.guide_frame.pack(fill=Tk.X)
        ok_cancel_frame = Tk.Frame(box)
        ok_cancel_frame.pack()

        sb_col = 0
        self.select_frame = ModelSelectFrame(standard_box, self, text="Model and Column Selection", labelanchor=Tk.N)
        self.select_frame.grid(row=0, column=0, padx=10)

        sb_col += 1
        self.advanced_frame = AdvancedFrame(standard_box, self)
        self.advanced_frame.grid(row=0, column=sb_col, pady=5)

        sb_col += 1
        self.preview_frame = PreviewButtonFrame(standard_box,
                                    dialog=self, use_elution_models=True, judge_holder=self.judge_holder,
                                    editor=self.get_current_frame())
        self.preview_frame.grid(row=0, column=sb_col, padx=10, pady=5, ipady=12)

        from molass_legacy.QuickAnalysis.AnalysisGuide import get_data_correction_state_text
        text = "Showing data processed with " + get_data_correction_state_text()
        guide_label = Tk.Label(self.guide_frame, text=text, bg='white')
        guide_label.pack(fill=Tk.X, pady=10)
        self.guide_message = guide_label

        w = Tk.Button(ok_cancel_frame, text="OK", width=10, command=self.ok, default=Tk.ACTIVE)
        w.grid(row=0, column=0, pady=5)
        self.ok_button = w

        space = Tk.Frame(ok_cancel_frame, width=160)
        space.grid(row=0, column=1)

        w = Tk.Button(ok_cancel_frame, text="Cancel", width=10, command=self.user_cancel)
        w.grid(row=0, column=2, pady=5)
        self.cancel_button = w

        if False:
            w = Tk.Button(additional_box, text="Reset from Log", width=12, command=self.show_reset_from_log_dialog)
            w.pack(padx=10, pady=5)
            self.restore_button = w

        # self.bind("<Return>", self.ok)
        self.bind("<Escape>", self.user_cancel)
        self.after(100, self.update_selection_state)

    def update_selection_state(self):
        self.select_frame.update_selection_state()

    def use_column_constraints_tracer(self, *args):
        state = Tk.DISABLED if self.use_column_constraints.get() == 0 else Tk.NORMAL
        self.selector.config(state=state)

    def change_model(self, selected_name):
        from .DecompEditorFrame import DecompEditorFrame
        current = self.current
        next_ = MODEL_NAMES.index(selected_name)
        if next_ == current:
            return

        self.advanced_frame.disable_buttons()
        self.update()   # to make the newly selected name appear timely

        next_frame = self.frames[next_]
        if next_frame.is_delayed():
            next_model = next_frame.model
            next_model.set_delayed_off()
            next_frame.destroy()
            new_frame = DecompEditorFrame(self.body_frame, self, self.sd, self.mapper, self.corbase_info, next_model)
            new_frame.pack()
            new_frame.pack_forget()
            self.frames[next_] = new_frame

        global_mode = self.global_fit.get()
        if global_mode:
            self.global_frames[current].pack_forget()
            global_frame = self.get_global_frame(next_)
            global_frame.pack()
        else:
            current_fig_type = self.frames[current].elution_fig_type
            self.frames[current].pack_forget()
            self.frames[next_].pack()
            self.frames[next_].refresh_frame(elution_fig_type=current_fig_type)
        self.current = next_
        self.preview_frame.update_modelframe()
        self.advanced_frame.update_mode_dependent_state()
        self.update_selection_state()

    def get_conc_depend(self):
        return self.preview_frame.get_conc_depend()

    def apply( self ):
        current_frame = self.frames[self.current]
        current_frame.apply()
        self.preview_frame.update_settings()        # concentration_datatype will be set_setting-ed here
        self.applied = True
        set_setting('use_elution_models', 1)
        set_setting('has_elution_models', True)
        set_setting('matrix_formulation', 1)
        self.try_image_save()

    def try_image_save( self ):
        # TODO: move this code to Tester.py

        test_pattern = get_setting('test_pattern')
        if test_pattern is None or test_pattern < 7:
            return

        fig_type = self.get_current_frame().elution_fig_type
        data_type = '_xray' if fig_type == 0 else '_uv'
        decomp_image_folder = get_setting('decomp_image_folder' + data_type )
        if decomp_image_folder is not None and os.path.exists(decomp_image_folder):
            self.save_the_figure( decomp_image_folder, get_setting('analysis_name') )

    def save_the_figure(self, folder, analysis_name):
        from pyautogui import screenshot
        from molass_legacy.KekLib.TkUtils import split_geometry

        self.focus_force()
        self.update()
        filename = analysis_name.replace( 'analysis', 'figure' ) + '.png'
        path = os.path.join( folder, filename )
        w, h, x, y = split_geometry(self.geometry())
        # print( path, w, h, x, y  )
        screenshot( path, region=(x, y, w, h) ) # doesn't support multi-screen

    def get_current_frame(self):
        return self.frames[self.current]

    def get_current_model(self):
        return self.get_current_frame().model

    def get_current_params_array(self):
        import numpy as np
        frame = self.get_current_frame()
        params_list = []
        for rec in frame.opt_recs:
            params = rec.get_params()
            params_list.append(params)
        return np.array(params_list)       

    def get_current_modelname(self):
        return self.get_current_frame().model_name

    def make_range_info(self, concentration_datatype):
        current_frame = self.get_current_frame()        # note that current frame will never be a dummy
        return current_frame.make_range_info(concentration_datatype)

    def get_applied_model(self):
        return self.frames[self.current].model

    def get_model_fit_errors(self):
        fit_errors = []
        for frame in self.frames:
            fit_errors.append( frame.get_fit_error() )
        return fit_errors

    def show_reset_from_log_dialog(self):
        from molass_legacy.Decomposer.DecompResetDialog import DecompResetDialog
        self.get_current_frame().log_info_for_reset(modification="currently active")
        dialog = DecompResetDialog(self)
        dialog.show()

    def global_fit_tracer(self, *args):
        global_mode = self.global_fit.get()
        current = self.current
        current_frame = self.get_current_frame()
        global_frame = self.get_global_frame(current)
        if global_mode:
            current_frame.pack_forget()
            global_frame.pack()
        else:
            global_frame.pack_forget()
            current_frame.pack()

    def get_global_frame(self, i):
        global_frame = self.global_frames[i]
        if global_frame is None:
            self.config(cursor='wait')
            self.update()
            decomp_result = self.frames[i].compute_global_fitting()
            global_frame = self.create_new_frame(i, decomp_result)
            self.global_frames[i] = global_frame
            self.config(cursor='')
        return global_frame

    def create_new_frame(self, i, decomp_result):
        from .DecompEditorFrame import DecompEditorFrame
        model = self.get_model_list()[i]
        frame = DecompEditorFrame(self.body_frame, self, self.sd, self.mapper, self.corbase_info, model, decomp_result=decomp_result, global_flag=True)
        return frame

    def update_global_btn_state(self, fig_type):
        # task: remove this method
        pass

    def update_current_frame_with_result(self, decomp_result, debug=False):
        self.get_current_frame().update_with_result(decomp_result, debug=debug)

    def reset_current_frame(self):
        self.get_current_frame().reset()



"""
    DenssUtils.py

    Copyright (c) 2019-2025, SAXS Team, KEK-PF

    Note that the original DENSS by T. Grant is licensed under GPL3.

    DENSS is available open source under the GPL 3.0 license.
    This means that anyone is free to download, utilize and develop the code,
    provided that any derivative code developed based on DENSS
    or its underlying algorithm is also distributed under the GPL 3.0 license
    and retains the original copyright.
"""
import numpy as np
import sys, argparse, os
import logging
import time
import molass_legacy.DENSS.denss as denss
import molass_legacy.DENSS.denss.options as dopts
from molass_legacy.KekLib.BasicUtils import Struct

MAXNUM_STEPS = 20000

def fit_data_impl(q, a, e, file=None, D=None, alpha=None, max_alpha=None, nes=2, extrapolate=True, gui=False, use_memory_data=False):
    Iq = np.vstack( [q, a, e] ).T

    # task: update this construction automatically from bin\denss.fit_data.py
    args = Struct(
        alpha = alpha,
        units = "a",
        max_alpha = max_alpha,
        output = "data_name",
        file = file,    # 
        dmax = D,
        nes = nes,
        n1 = None,
        n2 = None,
        ignore_errors = True,
        q = None,
        qmax = None,
        nq = None,
        r = None,
        max_dmax = None,
        qfile = None,
        rfile = None,
        extrapolate = extrapolate,
        nr = None,
        plot = True,
        log = True,
        write_shannon = True,
        )

    """
    DENSS strongly recommends to include I(q=0).
    To conform, we are using denss.fit_data.py here processing logic
    to modify the input data

    below is the extracted default processing from denss.fit_data.py
    lines[73:194] as of ver. 1.8.6
    """
### DENSS.bin.denss.fit_data.py copy & modify BEIGN ###
    alpha = args.alpha

    if args.output is None:
        fname_nopath = os.path.basename(args.file)
        basename, ext = os.path.splitext(fname_nopath)
        output = basename
    else:
        output = args.output

    if use_memory_data:
        # as invoked from LRF preview 
        Iq = np.array([q, a, e]).T
    else:
        if args.ignore_errors:
            Iq = np.genfromtxt(args.file, invalid_raise=False, usecols=(0, 1))
        else:
            Iq = np.genfromtxt(args.file, invalid_raise=False, usecols=(0, 1, 2))
    if len(Iq.shape) < 2:
        print("Invalid data format. Data file must have 3 columns: q, I, errors. Alternatively, disable errors with --ignore_errors option (sets errors to 1.0).")
        exit()
    if Iq.shape[1] < 3 or args.ignore_errors:
        print("WARNING: Only 2 columns given. Data should have 3 columns: q, I, errors.")
        print("WARNING: Setting error bars to 1.0 (i.e., ignoring error bars)")
        Iq2 = np.zeros((Iq.shape[0], 3))
        Iq2[:,:2] = Iq[:,:2]
        Iq2[:,2] += 1.0 #set error bars to 1.0
        Iq = Iq2
    Iq = Iq[~np.isnan(Iq).any(axis = 1)]
    #get rid of any data points equal to zero in the intensities or errors columns
    idx = np.where((Iq[:,1]!=0)&(Iq[:,2]!=0))
    Iq = Iq[idx]

    nes = args.nes

    if args.units == "nm":
        Iq[:,0] *= 0.1

    if args.n1 is None:
        n1 = 0
    else:
        n1 = args.n1
    if args.n2 is None:
        n2 = len(Iq[:,0])
    else:
        n2 = args.n2

    Iq_orig = np.copy(Iq)

    if args.dmax is None:
        #estimate dmax directly from data
        #note that denss.estimate_dmax does NOT extrapolate
        #the high q data, even though by default
        #denss.Sasrec does extrapolate.
        D, sasrec = denss.estimate_dmax(Iq, clean_up=True)
    else:
        D = args.dmax

    if args.max_dmax is None:
        args.max_dmax = 2.*D

    if args.rfile is not None:
        r = np.loadtxt(args.rfile,usecols=(0,))
    else:
        r = None

    if args.qfile is not None:
        qc = np.loadtxt(args.qfile,usecols=(0,))
    else:
        qc = None

    print("Dmax = %.2f"%D)
    qmax = Iq[n1:n2,0].max()
    if (qc is None) and (args.extrapolate):
        qmaxc = qmax*3.0
    elif (qc is None):
        qmaxc = qmax
    else:
        qmaxc = qc.max()

    #let a user set a desired set of q values to be calculated
    #based on a given qmax and nq
    if (args.qmax is not None) or (args.nq is not None):
        if args.qmax is not None:
            qmaxc = args.qmax
        else:
            qmaxc = qmaxc
        if args.nq is not None:
            nqc = args.nq
        else:
            nqc = 501
        qc = np.linspace(0,qmaxc,nqc)
        #assume if they are giving args.qmax or args.nq, that they want to
        #disable extrapolation
        args.extrapolate = False

    nsh = qmax/(np.pi/D)
    nshc = qmaxc/(np.pi/D)
    print("Number of experimental Shannon channels: %d"%(nsh))
    print("Number of calculated Shannon channels: %d"%(nshc))
    if (nsh > 500) or (nshc>500):
        print("WARNING: Nsh > 500. Calculation may take a while. Please double check Dmax is accurate.")
        #give the user a few seconds to cancel with CTRL-C
        waittime = 10
        try:
            for i in range(waittime+1):
                sys.stdout.write("\rTo cancel, press CTRL-C in the next %d seconds. "%(waittime-i))
                sys.stdout.flush()
                time.sleep(1)
            print()
        except KeyboardInterrupt:
            print("Canceling...")
            exit()


    #calculate chi2 when alpha=0, to get the best possible chi2 for reference
    sasrec = denss.Sasrec(Iq[n1:n2], D, qc=qc, r=r, nr=args.nr, ne=nes, alpha=0.0, extrapolate=args.extrapolate)
    ideal_chi2 = sasrec.calc_chi2()

    if args.alpha is None:
        alpha = sasrec.optimize_alpha(gui=gui)
    else:
        alpha = args.alpha
    sasrec = denss.Sasrec(Iq[n1:n2], D, qc=qc, r=r, nr=args.nr, alpha=alpha, ne=nes, extrapolate=args.extrapolate)

    #implement method of estimating Vp, Vc, etc using oversmoothing
    sasrec.estimate_Vp_etal()

### DENSS.bin.denss.fit_data.py copy  & modify  END ###

    work_info = Struct(Iq=Iq, args=args, alpha=alpha, n1=n1, n2=n2, Iq_orig=Iq_orig)
    return sasrec, work_info

def fit_data(q, a, e, D=None, extrapolate=False, return_sasrec=False):
    sasrec, work_info = fit_data_impl(q, a, e,
                                        D=D,
                                        alpha=0,        # backward compatibility
                                        max_alpha=None,
                                        extrapolate=extrapolate)
    if return_sasrec:
        return sasrec 
    else:
        return sasrec.qc, sasrec.Ic, sasrec.Icerr, sasrec.D

def fit_data_bc(q, a, e, extrapolate=False):    # backward compatible
    sasrec, work_info = fit_data_impl(q, a, e,
                                        D=100,
                                        alpha=0,
                                        max_alpha=10,
                                        extrapolate=extrapolate)
    return sasrec.qc, sasrec.Ic, sasrec.Icerr, sasrec.D

def get_dmax_with_datgnom(file_path):
    from molass_legacy.Saxs.Rdf import AtsasDatGnomDdf
    ddf = AtsasDatGnomDdf(file_path)
    dmax = ddf.guess_best_dmax()
    return dmax

def get_argparser(return_args=False):
    parser = argparse.ArgumentParser(description="DENSS: DENsity from Solution Scattering.\n A tool for calculating an electron density map from solution scattering data",
                                     formatter_class=argparse.RawTextHelpFormatter)
    if return_args:
        args = dopts.parse_arguments(parser, return_args=True)
        return args
    else:
        return parser

def run_denss_impl(qc, ac, ec, dmax, data_name, steps=MAXNUM_STEPS, progress_cb=None, use_gpu=False, gui=False):

    q = qc
    I = ac
    sigq = ec
    isfit = True    # it is assumed that the data is already fitted with fit_data_impl()

    sys.argv = ['dummy-script', '-f', data_name]
    if use_gpu:
        sys.argv += ['-gpu']
    parser = get_argparser()
    data_proxy = [q, I, sigq, I, dmax, isfit]
    args = dopts.parse_arguments(parser, data_proxy=data_proxy)

    qdata, Idata, sigqdata, qbinsc, Imean, chis, rg, supportV, rho, side, fit, final_chi2 = denss.reconstruct_abinitio_from_scattering_profile(
        # copied from denss/scripts/denss_abintio.py BEGIN
        q=args.q,
        I=args.I,
        sigq=args.sigq,
        dmax=args.dmax,
        qraw=args.qraw,
        Iraw=args.Iraw,
        sigqraw=args.sigqraw,
        ne=args.ne,
        voxel=args.voxel,
        oversampling=args.oversampling,
        recenter=args.recenter,
        recenter_steps=args.recenter_steps,
        recenter_mode=args.recenter_mode,
        positivity=args.positivity,
        positivity_steps=args.positivity_steps,
        extrapolate=args.extrapolate,
        output=args.output,
        steps=args.steps,
        ncs=args.ncs,
        ncs_steps=args.ncs_steps,
        ncs_axis=args.ncs_axis,
        ncs_type=args.ncs_type,
        seed=args.seed,
        support_start=args.support_start,
        shrinkwrap=args.shrinkwrap,
        shrinkwrap_old_method=args.shrinkwrap_old_method,
        shrinkwrap_sigma_start=args.shrinkwrap_sigma_start,
        shrinkwrap_sigma_end=args.shrinkwrap_sigma_end,
        shrinkwrap_sigma_decay=args.shrinkwrap_sigma_decay,
        shrinkwrap_threshold_fraction=args.shrinkwrap_threshold_fraction,
        shrinkwrap_iter=args.shrinkwrap_iter,
        shrinkwrap_minstep=args.shrinkwrap_minstep,
        chi_end_fraction=args.chi_end_fraction,
        write_xplor_format=args.write_xplor_format,
        write_freq=args.write_freq,
        enforce_connectivity=args.enforce_connectivity,
        enforce_connectivity_steps=args.enforce_connectivity_steps,
        enforce_connectivity_max_features=args.enforce_connectivity_max_features,
        cutout=args.cutout,
        quiet=args.quiet,
        # copied from denss/scripts/denss_abintio.py END
        gui=gui,
        DENSS_GPU=args.DENSS_GPU,
        progress_cb=progress_cb)

def run_denss_impl_dummy(qc, ac, ec, dmax, infile_name):
    import time
    print("\n Step     Chi2     Rg    Support Volume")
    print(" ----- --------- ------- --------------")

    for j in range(100):
        time.sleep(0.1)
        sys.stdout.write("\r% 5i " % (j))
        sys.stdout.flush()

def get_denssfolder(parent=None):
    from molass_legacy._MOLASS.SerialSettings import get_setting
    from molass_legacy.KekLib.BasicUtils import mkdirs_with_retry

    analysis_folder = get_setting('analysis_folder')
    if analysis_folder is None or analysis_folder == "":
        return None

    denss_folder = (analysis_folder + '/DENSS').replace('\\', '/')
    if not os.path.exists(denss_folder):
        mkdirs_with_retry(denss_folder)

    return denss_folder

def get_outfolder(job_id=None, parent=None):
    denss_folder = get_denssfolder(parent=parent)
    if denss_folder is None:
        return None

    import re
    if job_id is None:
        out_folder = denss_folder + '/000'
        while True:
            if os.path.exists(out_folder):
                out_folder = re.sub(r'/(\d+)$', lambda m: '/%03d' % (int(m.group(1)) + 1), out_folder)
            else:
                break
    else:
        out_folder_init = denss_folder + '/%03d' % job_id
        out_folder = out_folder_init
        i = 0
        while True:
            if os.path.exists(out_folder):
                i += 1
                out_folder =  out_folder_init + '_%d' % i
            else:
                break

    return out_folder

def get_denss_log_items(path):
    import re
    item_re = re.compile(r'^\d+:\d+:\d+\s+Final\s+(.+):\s+(.+)$')
    ret_dict = {}
    with open(path) as log:
        for line in log:
            m = item_re.match(line)
            if m:
                ret_dict[m.group(1)] = float(m.group(2))
    return ret_dict

def run_pdb2mrc(in_file, queue=None):
    import time
    from molass_legacy.KekLib.SubProcess import Popen   # suppresses the child process window.
    from molass_legacy.KekLib.BasicUtils import get_home_folder
    print("Generating an mrc file.")
    script_path = os.path.join(get_home_folder(), r'molass_legacy\DENSS\denss\scripts\denss_pdb2mrc.py')
    python  = sys.executable.replace('pythonw.exe', 'python.exe')       # running with pythonw.exe seems inappropriate
    out_file = in_file.replace('.pdb', '')
    cmd = [python, script_path, '-f', in_file, '-o', out_file]
    # print(cmd)

    """
        currently not receiving the stdout of the child process.
        TODO: better print the progress.
    """
    p = Popen(cmd)
    while p.poll() is None:
        time.sleep(1)
        if queue is not None:
            queue.put((1, '.'))
    returncode = p.poll()

    if queue is not None:
        queue.put((1, '\n'))

    if returncode == 0:
        out_file = out_file + ".mrc"
        print("Generated %s" %  out_file)
    else:
        if queue is not None:
            queue.put((1, ret.stderr.decode() ))
        out_file = None
        print("Failed with returncode: %d" % returncode)

    return out_file



# coding: utf-8
"""
    FlowSimulator.py

    Copyright (c) 2019, SAXS Team, KEK-PF
"""
import os
import numpy as np
from bisect import bisect_left
import logging
from openpyxl import load_workbook
from MethodFile import MethodFile

#                    D, F, H, L
TOTAL_RATE_COLUMNS = [3, 5, 7, 11]
RATE_B_START_COLUMN = 5
RATE_B_STOP_COLUMN = 6
SAFE_MARGIN_RATIO = 0.05

class SimInfo:
    def __init__(self, **kwargs): 
        self.__dict__.update(kwargs)

class FlowSimulator(MethodFile):
    def __init__(self, filepath, bookpath=None, num_files=None):
        MethodFile.__init__(self, filepath)
        self.logger = logging.getLogger(__name__)
        if bookpath is None:
            bookpath = filepath.replace('.mtd', '.xlsx')
        self.bookpath = bookpath
        self.num_files = num_files

        self.opt_factor = None
        self.start_duration = None
        self.start_frate_total = None
        self.start_time = None
        self.start_conc = None
        self.stop_duration = None
        self.stop_frate_total = None
        self.stop_time = None
        self.stop_conc = None
        self.excel_book_ok = False
        self.eno_interval = None

        try:
            self.read_the_book(bookpath)
            self.excel_book_ok = True
        except FileNotFoundError:
            pass
        except:
            from molass_legacy.KekLib.ExceptionTracebacker import ExceptionTracebacker
            etb = ExceptionTracebacker()
            print(etb)

    def read_the_book(self, bookpath):
        debug = True
        self.wb = load_workbook(bookpath, data_only=True)
        self.delay_factor = self.wb['Parameter']['B11'].value
        self.eno_interval = self.wb['Time vs conc.']['B3'].value
        self.time_delta = self.eno_interval/2
        if debug:
            print('delay_factor=', self.delay_factor, 'eno_interval=', self.eno_interval)

    def set_params(self, delay_factor=None, debug=False):
        if not self.excel_book_ok:
            self.set_params_without_book(delay_factor)
            return

        if delay_factor is None:
            delay_factor = self.delay_factor
        ws = self.wb['Method']
        range_ = ws['A3:O30']
        for k, row in enumerate(range_, start=1):
            values = [cell.value for cell in row]
            if debug:
                print([k], values)
            name = values[0]
            if name is None:
                break

            if name == 'const flow rate (pump A, B)':
                self.start_duration = values[-1]
            elif name == 'Total flow rate modulation':
                self.pre_start_time = values[-1]
            elif name == 'Y channel initialize':
                frate_total = np.sum([values[k] for k in TOTAL_RATE_COLUMNS])
                self.y_channel_time = values[-1] + delay_factor/frate_total
            elif name == 'Linearly changing flow rate':
                self.stop_duration = values[-1]
                self.start_frate_total = np.sum([values[k] for k in TOTAL_RATE_COLUMNS])
                self.start_time = self.start_duration + delay_factor/self.start_frate_total
                self.start_conc = values[RATE_B_START_COLUMN]/self.start_frate_total
                rate_b_stop = values[RATE_B_STOP_COLUMN]
            elif name == 'const. uL/min (pump B)':
                self.stop_frate_total = np.sum([values[k] for k in TOTAL_RATE_COLUMNS])
                self.stop_time = self.stop_duration + delay_factor/self.stop_frate_total
                self.stop_conc = rate_b_stop/self.stop_frate_total
            last_values = values

        self.end_time = last_values[-1] + delay_factor/np.sum([last_values[k] for k in TOTAL_RATE_COLUMNS])

        if self.num_files is not None:
            self.verify_eno_interval(self.num_files)

        if debug:
            for r in [  self.start_duration,
                        self.pre_start_time,
                        self.y_channel_time,
                        self.start_frate_total,
                        self.stop_duration,
                        self.stop_frate_total,
                        self.start_time,
                        self.stop_time,
                        self.start_conc,
                        self.stop_conc,
                        ]:
                print(r)

    def verify_eno_interval(self, num_files):
        ratio = num_files/self.end_time
        a_diff = abs(ratio - 1)
        ratio_xls = num_files*self.eno_interval/self.end_time
        a_diff_xls = abs(ratio_xls - 1)
        if a_diff_xls > a_diff:
            # TODO: this fix should be done earlier in self.set_params by giving num_files to it
            fix_value = 1.0
            self.logger.warning("Excel book eno_interval %g has been changed to %g due to time scale inconsistency.", self.eno_interval, fix_value)
            self.eno_interval = fix_value

    def set_params_without_book(self, delay_factor):
        from MethodFileSimulator import MethodFileSimulator
        MethodFileSimulator.set_params(self, delay_factor, default_init=True, debug=True)

    def get_y_channel_interval(self, t):
        i = bisect_left(t, self.pre_start_time)
        j = bisect_left(t, self.start_time)
        return i, j

    def get_linear_range_info(self, t, vscale=None):
        i = bisect_left(t, self.pre_start_time)
        c = bisect_left(t, self.y_channel_time)
        j = bisect_left(t, self.start_time)
        k = bisect_left(t, self.stop_time)
        print('t[[i, j, k]]=', t[[i, j, k]])
        return SimInfo(start=i, stop=None, yc_peak=c, lin_start=j, lin_end=k, factor=self.opt_factor, vscale=vscale)

    def get_y_channel_index(self, t):
        return bisect_left(t, self.y_channel_time)

    def get_optimizer_indeces(self, t):
        i = bisect_left(t, self.y_channel_time)
        j = bisect_left(t, self.start_time)
        k = bisect_left(t, self.stop_time)
        return i, j, k

    def make_simulation_data(self, delay_factor=None, debug=False):
        if delay_factor is None:
            delay_factor = self.delay_factor

        self.opt_factor = delay_factor

        self.set_params(delay_factor, debug=debug)

        tsec = 0
        tmin = None
        sm_rows = []
        for n, dur, a1, a2, b1, b2, c1, c2, d1, d2, e1, e2 in self.rows:
            tr1 = a1 + b1 + c1 + e1
            delay = 0  if n < 3 else delay_factor/tr1   # to avoid obscure appearance
            tmin = tsec/60 + delay
            if tsec > 0:
                sm_rows[-1][-2] = tmin
            tr2 = a2 + b2 + c2 + e2
            sm_rows.append([n, tr1, tmin, b1/tr1])
            sm_rows.append([n, tr2, None, b2/tr2])
            tsec += dur

        sm_rows[-1][-2] = tmin

        if debug:
            for k, row in enumerate(sm_rows):
                print([k], row)

        sm_array = np.array(sm_rows)
        t = sm_array[:,2]
        print('eno_interval=', self.eno_interval)
        x = (t - t[0]+self.time_delta)/self.eno_interval - 1      # int(x) = eno
        y = sm_array[:,3]
        return t, x, y

    def guess_range(self, num_files=None):
        safe_margin = (self.stop_time - self.start_time)*SAFE_MARGIN_RATIO
        start = int((self.start_time + safe_margin)/self.eno_interval)
        stop  = int((self.stop_time - safe_margin)/self.eno_interval)+1

        if self.excel_book_ok:
            # self.eno_interval should have been verified
            pass
        else:
            ratio = num_files/self.end_time
            if abs(ratio - 1) > 0.1:
                self.eno_interval = 1/ratio
                start = int(start*ratio)
                stop = int(stop*ratio)
                self.logger.warning("Default eno_interval has been changed to %g due to time scale inconsistency.", self.eno_interval)

        return slice(start, stop)



"""
    Stochastic.ColumnSimulation.py

    Copyright (c) 2024, SAXS Team, KEK-PF
"""
import numpy as np
from matplotlib.patches import Rectangle, Circle, Wedge
from matplotlib.animation import FuncAnimation

solvant_color = "lightcyan"

class SolidGrain:
    def __init__(self, id_, center, radius, poreradius, poredist):
        self.id_ = id_
        self.center = np.asarray(center)
        self.radius = radius
        self.poreradius = poreradius
        self.poredist = poredist
        self.compute_poreentries()

    def compute_poreentries(self):
        entry_edges = []
        for dy in [-self.poredist, 0, self.poredist]:
            for y in [dy - self.poreradius, dy + self.poreradius]:
                dx = np.sqrt(self.radius**2 - y**2)
                for x in [-dx, dx]:
                    entry_edges.append(np.arctan2(y, x))
        entry_edges = sorted(entry_edges)
        entries = []
        entries.append((entry_edges[-1], entry_edges[0]))
        for k in range(len(entry_edges)//2-1):
            entries.append((entry_edges[2*k+1], entry_edges[2*k+2]))
        self.entries = np.array(entries)

    def draw_entries(self, ax):
        for entry in self.entries:
            points = np.array([self.center + self.radius*np.array([np.cos(r), np.sin(r)]) for r in entry])
            ax.plot(*points.T)

    def draw(self, ax, color=None, alpha=1):
        p = Circle(self.center, self.radius, color=color, alpha=alpha)
        ax.add_patch(p)
        cx, cy = self.center
        for dy in [-self.poredist, 0, self.poredist]:
            p = Rectangle((cx-self.radius, cy-self.poreradius+dy), self.radius*2, self.poreradius*2, color=solvant_color)
            ax.add_patch(p)
        p = Rectangle((cx-self.poreradius, cy-self.radius*0.9), self.poreradius*2, self.radius*1.8, color=color)
        ax.add_patch(p)

    def get_point_from_angle(self, angle):
        return self.center + np.array([np.cos(angle), np.sin(angle)])*self.radius

    def get_entry_including(self, angles, debug=False):
        if angles[0] > angles[1]:
            angles = np.flip(angles)
        w = np.where(np.logical_and(self.entries[:,0] < angles[0], angles[1] < self.entries[:,1]))[0]
        if debug:
            print('get_entry_including: ', self.id_, angles, w)
        if len(w) == 0:
            return None
        else:
            return w[0]

    def compute_bounce_vector(self, particle):
        pass

    def compute_inpore_nextpos(self, particle):
        pass

class Particle:
    def __init__(self, center, radius):
        self.center = np.asarray(center)
        self.radius = radius

    def draw(self, ax, color=None, alpha=1):
        p = Circle(self.center, self.radius, color=color, alpha=alpha)
        ax.add_patch(p)

    def enters_stationary(self, grain, last_particle=None, return_point_info=False, ax=None, debug=False):
        from molass_legacy.KekLib.CircleGeometry import get_intersections, circle_line_segment_intersection
        if self.radius >= grain.poreradius:
            if debug:
                # print("self.radius(%g) >= grain.poreradius(%g)" % (self.radius, grain.poreradius))
                pass
            return None

        ret = get_intersections(*self.center, self.radius, *grain.center, grain.radius)
        if debug:
            print("enters_stationary (1): ", grain.id_, ret, self.center, grain.center)
        if False:
            import molass_legacy.KekLib.DebugPlot as plt
            print("self.center=", self.center, "self.radius=", self.radius)
            with plt.Dp():
                fig, ax = plt.subplots()
                ax.set_title("enters_stationary debug")
                ax.set_aspect("equal")
                grain.draw(ax, alpha=0.5)
                self.draw(ax)
                ax.plot(*self.center, "o", color="black", markersize=1)
                fig.tight_layout()
                reply = plt.show()
                assert reply

        if ret is None:
            if last_particle is None:
                return None
            intersections = circle_line_segment_intersection(grain.center, grain.radius, self.center, last_particle.center, full_line=False)
            for point in intersections:
                tp = Particle(point, self.radius)
                tp_ret = tp.enters_stationary(grain)
                if debug:
                    print("tp_ret=", tp_ret)
                if tp_ret is not None:
                    if ax is not None:
                        ax.plot(point[0], point[1], "o", color="yellow")
                    return tp_ret
            return None
        
        if debug:
            print("enters_stationary (2): ", grain.id_)

        v1 = np.asarray(ret[0]) - grain.center
        v2 = np.asarray(ret[1]) - grain.center
        angles = []
        for v in [v1,v2]:
            r = np.arctan2(v[1], v[0])      # np.atan2(y, x)
            angles.append(r)

        if return_point_info:
            return angles, ret

        i = grain.get_entry_including(angles)
        if i is None:
            return None

        return angles, ret, i

    def stationary_move(self, grain, last_px, last_py, px, py, debug=False):

        def get_next_position():
            from importlib import reload
            import Stochastic.StationaryMove as sm
            reload(sm)
            from Stochastic.StationaryMove import get_next_position_impl
            debug_info = (fig, ax) if debug else None
            nx, ny, state = get_next_position_impl(self, grain, last_px, last_py, px, py, debug_info=debug_info)
            if debug:
                fig.canvas.draw_idle()
            return nx, ny, state

        if debug:
            import molass_legacy.KekLib.DebugPlot as plt
            extra_button_specs = [
                ("Next", get_next_position),
            ]
            with plt.Dp(extra_button_specs=extra_button_specs):
                fig, ax = plt.subplots()
                ax.set_title("stationary_move debug")
                ax.set_aspect("equal")
                grain.draw(ax)
                dx = px - last_px
                dy = py - last_py
                vlen = np.sqrt(dx**2 + dy**2)
                ax.arrow(x=last_px, y=last_py, dx=dx, dy=dy, width=0.0005, head_width=0.002, length_includes_head=True,
                         head_length=0.2*vlen, color='black', alpha=0.5)
                self.draw(ax, alpha=0.5)
                cx, cy = grain.center
                r = grain.radius * 1.2
                ax.set_xlim(cx - r, cx + r)
                ax.set_ylim(cy - r, cy + r)
                fig.tight_layout()
                reply = plt.show()
                assert reply

        return get_next_position()

class NewGrain(SolidGrain):
    def __init__(self, id_, center, radius, num_pores):
        self.id_ = id_
        self.center = center
        self.radius = radius
        self.num_pores = num_pores
        self.poreradius = np.pi*radius/(2*num_pores)
        print("poreradius=", self.poreradius)
        self.x = np.ones(num_pores*2)
        self.colors = ['lavender', 'gray'] * num_pores
        self.compute_poreentries()

    def compute_poreentries(self):
        unit_angle = 2*np.pi/self.num_pores
        half_angle = unit_angle/2
        entries = []
        wedge_rad_pairs = []
        for i in range(self.num_pores):
            angle = i*unit_angle
            entries.append((angle, angle+half_angle))
            wedge_rad_pairs.append((angle, angle+half_angle))
            wedge_rad_pairs.append((angle+half_angle, angle+half_angle*2))
        self.entries = np.array(entries)
        self.wedge_rad_pairs = wedge_rad_pairs

    def draw(self, ax):
        # task: use patches.Wedge instead
        # ax.pie(self.x, colors=self.colors, radius=self.radius, center=self.center)
        draw_wedges(ax, self.center, self.radius, self.wedge_rad_pairs, self.colors)

def new_grain_unit_test():
    import molass_legacy.KekLib.DebugPlot as plt

    radius = 0.2
    num_pores = 10
    colors = ['gray', 'pink'] * num_pores
    print("entry lenght=", radius*np.pi/num_pores)
    grain = NewGrain((0, 0), (0.5, 0.5), radius, num_pores)
    p0 = Particle((0.33, 0.7), 0.05)
    p1 = Particle((0.35, 0.65), 0.05)
    p2 = Particle((0.7, 0.7), 0.025)
    p3 = Particle((0.64, 0.64), 0.025)
    p4 = Particle((0.75, 0.55), 0.025)
    p5 = Particle((0.65, 0.52), 0.025)
    p6 = Particle((0.65, 0.33), 0.025)
    p7 = Particle((0.57, 0.36), 0.025)

    with plt.Dp():
        fig, ax = plt.subplots(figsize=(5,5))
        ax.set_aspect('equal')
        back = Rectangle((0.1, 0), 0.8, 1, color=solvant_color)
        ax.add_patch(back)
        grain.draw(ax)

        last_particle = None
        for k, particle in enumerate([p0, p1, p2, p3, p4, p5, p6, p7]):
            i, j = divmod(k,2)
            if j == 0:
                last_particle = None
            particle.draw(ax, alpha=0.3, color='C%d' % i)
            ret = particle.enters_stationary(grain, last_particle=last_particle, ax=ax)
            print([k], ret)
            if ret is not None:
                angles, points, i = ret
                for angle, point in zip(angles, points):
                    ax.plot(*point, "o", color="yellow")
                    ax.plot(*grain.get_point_from_angle(angle), "o", color="red", markersize=3)
            last_particle = particle

        grain.draw_entries(ax)

        ax.set_xlim(0,1)
        ax.set_ylim(0,1)
        fig.tight_layout()
        plt.show()

def grain_particle_test():
    import molass_legacy.KekLib.DebugPlot as plt

    grain  = SolidGrain((0,0), (0.5,0.5), 0.2, 0.03, 0.12)
    p0 = Particle((0.33, 0.7), 0.05)
    p1 = Particle((0.35, 0.65), 0.05)
    p2 = Particle((0.7, 0.7), 0.025)
    p3 = Particle((0.67, 0.63), 0.025)
    p4 = Particle((0.75, 0.5), 0.025)
    p5 = Particle((0.65, 0.5), 0.025)
    p6 = Particle((0.62, 0.3), 0.025)
    p7 = Particle((0.6, 0.38), 0.025)

    with plt.Dp():
        fig, ax = plt.subplots(figsize=(5,5))
        ax.set_aspect('equal')

        back = Rectangle((0.1, 0), 0.8, 1, color=solvant_color)
        ax.add_patch(back)

        grain.draw(ax, color="gray")

        last_particle = None
        for k, particle in enumerate([p0, p1, p2, p3, p4, p5, p6, p7]):
            i, j = divmod(k,2)
            if j == 0:
                last_particle = None
            particle.draw(ax, alpha=0.3, color='C%d' % i)
            ret = particle.enters_stationary(grain, last_particle=last_particle, ax=ax)
            print([k], ret)
            if ret is not None:
                angles, points, i = ret
                for angle, point in zip(angles, points):
                    ax.plot(*point, "o", color="yellow")
                    ax.plot(*grain.get_point_from_angle(angle), "o", color="red", markersize=3)
            last_particle = particle

        grain.draw_entries(ax)

        fig.tight_layout()
        plt.show()

def draw_wedges(ax, center, radius, rad_pairs, colors):
    scale = 180/np.pi
    for (r1, r2), c in zip(rad_pairs, colors):
        wedge = Wedge(center, radius, scale*r1, scale*r2, color=c)
        ax.add_patch(wedge)

def demo(debug=False):
    import molass_legacy.KekLib.DebugPlot as plt
    from matplotlib.widgets import Button

    save_gif = False
    smart_phone = False
    old_grain = False
    enbale_slice_drawing = True

    ymin, ymax = 0, 1
    xmin, xmax = 0.3, 0.7

    xm = 0.01
    ym = 0.03
    rs = 0.06
    if old_grain:
        pyshift = 0.03
        rp = 0.009
    else:
        num_pores = 32

    circle_cxv = np.linspace(xmin, xmax, 7)
    circle_cyv = np.flip(np.linspace(ymin+ym+rs, ymax-ym-rs, 8))
    cxx, cyy = np.meshgrid(circle_cxv, circle_cyv)
    if debug:
        print("cxx=", cxx)
        print("cyy=", cyy)

    psizes = np.array([8, 3, 1])
    markersizes = np.array([8, 4, 2])
    pcolors = ["green", "blue", "red"]
    ptype_indeces = np.array(list(np.arange(len(psizes)))*200)
    np.random.shuffle(ptype_indeces)
    num_particles = len(ptype_indeces)
    grain_references = -np.ones(num_particles, dtype=int)

    init_pxv = np.linspace(xmin+0.01, xmax-0.01, num_particles)
    init_pyv = np.ones(num_particles)*ymax

    num_frames = 400
    sigma = ymax/num_frames*2
    du = sigma
    particle_scale = 1/1000  # [10, 5, 1] => [0.01, 0.005, 0.001]
    radius_map = psizes*particle_scale
    print("radius_map=", radius_map)    
    radiusv = np.array([radius_map[i] for i in ptype_indeces])
    if debug:
        print("radiusv=", radiusv)
    rv = radiusv + rs

    with plt.Dp():
        if smart_phone:
            fig, ax1 = plt.subplots(figsize=(8,20))
            fig.suptitle("Size Exclusion Chromatography", fontsize=16, y=0.99)
        else:
            fig, (ax1, ax2) = plt.subplots(ncols=2, figsize=(16,20))
            fig.suptitle("Size Exclusion Chromatography Illustrative Animation", fontsize=24, y=0.99)
        main_fig = fig

        pause = False
        def on_click(event):
            nonlocal pause
            print('on_click')
            if event.inaxes != ax1:
                return
            pause ^= True
            
        fig.canvas.mpl_connect('button_press_event', on_click)

        button_ax = fig.add_axes([0.85, 0.05, 0.1, 0.03])
        def draw_slice_states(event):
            from Stochastic.ColumnSliceStates import draw_slice_states_impl
            print("draw_slice_states")
            if event.inaxes != button_ax:
                return
            draw_slice_states_impl(fig, ax2, grains, pxv, pyv, inmobile_states)

        debug_btn = Button(button_ax, 'Draw Slice States', hovercolor='0.975')
        debug_btn.on_clicked(draw_slice_states)

        def plot_column_structure(ax):
            ax.set_axis_off()

            p = Rectangle(
                        (xmin, ymin),      # (x,y)
                        xmax - xmin,          # width
                        ymax - ymin,    # height
                        facecolor   = solvant_color,
                        # alpha       = 0.2,
                    )
            ax.add_patch(p)

            def plot_circles(radius, color="gray", alpha=1, create_grains=False, draw_rectangle=False, debug=False):
                if create_grains:
                    grains = []
                
                for i, y in enumerate(circle_cyv):
                    for j, x in enumerate(circle_cxv):
                        if i%2 == 0:
                            if j%2 == 0:
                                continue
                        else:
                            if j%2 == 1:
                                continue

                        if create_grains:
                            if old_grain:
                                grain = SolidGrain((i, j), (x, y), radius, rp, pyshift)
                            else:
                                grain = NewGrain((i, j), (x, y), radius, num_pores)

                        if draw_rectangle:
                            p = Rectangle((x-radius, y-radius*4.5), radius*2, radius*9, color=color, alpha=alpha)
                        else:
                            p = Circle((x, y), radius, color=color, alpha=alpha)
                            if create_grains:
                                if old_grain:
                                    grain = SolidGrain((i, j), (x, y), radius, rp, pyshift)
                                else:
                                    grain = NewGrain((i, j), (x, y), radius, num_pores)
                                if debug:
                                    print("create_grains: ", (i, j), (x, y))
                                    grain.draw_entries(ax)
                                grains.append(grain)
                        ax.add_patch(p)

                if create_grains:
                    return grains

            grains = plot_circles(rs, create_grains=True)

            if old_grain:
                for y in np.linspace(ymin+ym+rs, ymax-ym-rs, 8):
                    for dy in [-pyshift, 0, pyshift]:
                        y_ = y + dy
                        p = Rectangle(
                                (xmin, y_-rp),   # (x,y)
                                xmax - xmin,    # width
                                rp*2,           # height
                                facecolor   = solvant_color,
                                # alpha       = 0.2,
                            )
                        ax.add_patch(p)
                plot_circles(rs*0.2, draw_rectangle=True)
                plot_circles(rs, color="pink", alpha=0.5)
            else:
                for grain in grains:
                    grain.draw(ax)

            """
            set_xlim or set_ylim must be called after plot_circles maybe due to NewGrain.graw which calles ax.pie
            """
            if not smart_phone:
                ax.set_aspect('equal')
            ax.set_xlim(xmin, xmax)    
 
            ax.set_ylim(0,1)

            return grains

        grains = plot_column_structure(ax1)
        xxv = []
        yyv = []
        for grain in grains:
            x, y = grain.center
            xxv.append(x)
            yyv.append(y)
        xxv = np.array(xxv)
        yyv = np.array(yyv)

        particles = []
        for k, x in enumerate(init_pxv):
            m = ptype_indeces[k]
            particle, = ax1.plot(x, ymax, "o", markersize=markersizes[m], color=pcolors[m])
            particles.append(particle)

        fig.tight_layout()

        inmobile_states = np.ones(num_particles, dtype=bool)
        pxv = init_pxv.copy()
        pyv = init_pyv.copy()

        def touchable_indeces(inmobile_states, last_pxv, last_pyv, debug=False):
            indeces = []
            bounce_scales = []
            for k, (mobile, x, y) in enumerate(zip(inmobile_states, pxv, pyv)):
                distv = rv[k] - np.sqrt((xxv - x)**2 + (yyv - y)**2)
                w = np.where(distv > 0)[0]
                if len(w) == 0:
                    inmobile_states[k] = True
                    grain_references[k] = -1
                else:
                    j = w[0]
                    # print("w=", w, "j=", j, "distv=", distv[j], "x,y=", x, y)
                    grain = grains[j]
                    last_particle = Particle((last_pxv[k], last_pyv[k]), radiusv[k])
                    this_particle = Particle((x, y), radiusv[k])
                    ret = this_particle.enters_stationary(grain, last_particle=last_particle, debug=debug)
                    if mobile:
                        if ret is None:
                            indeces.append((k, j))
                            bounce_scales.append(distv[j])
                            inmobile_states[k] = True
                            grain_references[k] = -1
                        else:
                            inmobile_states[k] = False
                            grain_references[k] = j
                    else:
                        # ret = this_particle.exits_stationary(grain, last_particle=last_particle, debug=debug)
                        # task: restrict stationary move
                        pass
                    
            if len(indeces) == 0:
                return None
            
            touchables, porous_indeces = np.array(indeces, dtype=int).T
            if debug:
                print("(1) touchables=", touchables)
                print("(1) inmobile_states=", ''.join(map(lambda b: '%d' % b, inmobile_states)))
                print("(1) staying_grains =", ''.join(map(lambda j: '.' if j < 0 else chr(97+j), grain_references)))
            bounce_scales = np.array(bounce_scales)
            dxv = pxv[touchables] - xxv[porous_indeces]
            dyv = pyv[touchables] - yyv[porous_indeces]
            dlenv = np.sqrt(dxv**2 + dyv**2)
            scale = bounce_scales/dlenv*2
            bxv = dxv*scale
            byv = dyv*scale
            return touchables, np.array([bxv, byv]).T
    
        cancel_debug = False

        def compute_next_positions(debug=False):
            nonlocal pxv, pyv, cancel_debug
            if debug:
                print("(2) inmobile_states=", ''.join(map(lambda b: '%d' % b, inmobile_states)))
            last_pxv = pxv.copy()
            last_pyv = pyv.copy()
            dxv, dyv = np.random.normal(0, sigma, (2,num_particles))
            pxv += dxv
            pyv += dyv
            pyv[inmobile_states] -= du
            ret = touchable_indeces(inmobile_states, last_pxv, last_pyv)
            if ret is not None:
                # modify mobile move
                touchables, bounce_vects = ret
                pxv[touchables] += bounce_vects[:,0]
                pyv[touchables] += bounce_vects[:,1]

            # modify statinary move
            stationary_indeces = np.where(np.logical_not(inmobile_states))[0]
            # if not old_grain:
            for i in stationary_indeces:
                particle = Particle((pxv[i], pyv[i]), radiusv[i])
                grain = grains[grain_references[i]]
                nx, ny, state = particle.stationary_move(grain, last_pxv[i], last_pyv[i], pxv[i], pyv[i], debug=False)
                pxv[i] = nx
                pyv[i] = ny
                inmobile_states[i] = state

            exceed_left = pxv < xmin
            pxv[exceed_left] += -2*dxv[exceed_left]
            exceed_right = pxv > xmax
            pxv[exceed_right] += -2*dxv[exceed_right]

            exceed_top = pxv > ymax
            pyv[exceed_top] += -2*dyv[exceed_top]
            if debug and not cancel_debug:
                print("ret=", ret)
                with plt.Dp():
                    fig, ax = plt.subplots(figsize=(9,9))
                    plot_column_structure(ax)
                    U = pxv - last_pxv
                    V = pyv - last_pyv
                    ax.quiver(last_pxv, last_pyv, U, V, width=0.002,
                              angles='xy', scale_units='xy', scale=1, color="blue")

                    if ret is not None:
                        X = pxv[touchables] - bounce_vects[:,0]
                        Y = pyv[touchables] - bounce_vects[:,1]
                        U = 2*bounce_vects[:,0]
                        V = 2*bounce_vects[:,1]
                        ax.quiver(X, Y, U, V, width=0.002,
                              angles='xy', scale_units='xy', scale=1, color="red")

                    fig.tight_layout()
                    reply = plt.show()
                    if not reply:
                        cancel_debug = True
            if debug:
                print("(3) inmobile_states=", ''.join(map(lambda b: '%d' % b, inmobile_states)))
            return pxv, pyv

        def animate(i):
            if not pause:
                compute_next_positions()
            for k, p in enumerate(particles):
                p.set_data(pxv[k:k+1], pyv[k:k+1])
            return particles

        def init():
            nonlocal pxv, pyv, rv
            pxv = init_pxv.copy()
            pyv = init_pyv.copy()
            np.random.shuffle(ptype_indeces)
            radiusv = np.array([radius_map[i] for i in  ptype_indeces])
            if debug:
                print("init: radiusv=", radiusv)
            rv = radiusv + rs
            return animate(0)

        anim = FuncAnimation(fig, animate, init_func=init,
                                frames=num_frames, interval=100, blit=True)

        if save_gif:
            anim.save("sec.gif")
        plt.show()

if __name__ == "__main__":
    import sys
    sys.path.append("../lib")
    # new_grain_unit_test()
    # grain_particle_test()
    demo()


"""
    Solvers.NcallsEstimator.py

    Copyright (c) 2024, SAXS Team, KEK-PF
"""
def bh_estimator(niter, ncallbacks, latest_ncalls):
    pass

def ultranest_estimator(niter, ncallbacks, latest_ncalls):
    pass

ESTIMATIOR_FUNC_DICT ={
    "bh"        : bh_estimator,
    "ultranest" : ultranest_estimator,
}

def estimate_ncalls(solver_name, niter, ncallbacks, latest_ncalls):
    func = ESTIMATIOR_FUNC_DICT[solver_name]
    return func(niter, ncallbacks, latest_ncalls)


# coding: utf-8
"""
    Optimizer.WFM.py

    Copyright (c) 2020, SAXS Team, KEK-PF
"""
import numpy as np

def proof():
    A = np.random.uniform(0, 1, (4,2))
    B = np.random.uniform(0, 1, (2,3))
    C = np.random.uniform(0, 1, (4,2))
    D = np.random.uniform(0, 1, (2,3))
    pw = np.random.uniform(0, 1, (4,1))
    cw = np.random.uniform(0, 1, (1,3))

    W = pw@cw

    print('---- (1)')
    print(pw*A)
    print(A*pw)
    print('---- (2)')
    print((pw*A)@(B*cw))
    print(pw*(A@B)*cw)
    print('---- (3)')
    print(pw*(A - C))
    print(pw*A - pw*C)
    print('---- (4)')
    print((B + D)*cw)
    print(B*cw + D*cw)



"""
    PdbCrysol.py

    Copyright (c) 2022, SAXS Team, KEK-PF
"""
import sys
import os
import numpy as np

def pdb2crysol_impl(pdbid, out_folder):
    from Pdb.PypdbLite import get_pdb_file
    from molass_legacy.ATSAS.Crysol import CrysolExecutor

    ret = get_pdb_file(pdbid, filetype='pdb')
    pdb_file = pdbid + '.pdb'
    pdb_path = os.path.join(out_folder, pdb_file)
    with open(pdb_path, "w") as fh:
                        fh.write(ret)

    crysol = CrysolExecutor()
    print('crysol.execute')
    crysol.execute(pdb_path)
    # crysol.move(path, sas_folder)

if __name__ == '__main__':
    sys.path.append(os.path.dirname(os.path.abspath(__file__)) + '/..')
    import molass_legacy.KekLib, SerialAnalyzer

    # pdb2crysol_impl("1AON", "temp")
    # pdb2crysol_impl("1FFK", "temp")
    # pdb2crysol_impl("1GQ2", "temp")
    # pdb2crysol_impl("1HTO", "temp")
    # pdb2crysol_impl("1VQ9", "temp")
    # pdb2crysol_impl("1XI4", "temp")
    # pdb2crysol_impl("2J28", "temp")
    pdb2crysol_impl("2RDO", "temp")



"""
    SerialSettings.py

    Copyright (c) 2016-2024, SAXS Team, KEK-PF
"""
import os
from molass_legacy.KekLib.PersistentPy import PersistentPy, get_py_file_path
try:
    from molass_legacy._MOLASS.Version import get_version_string
except:
    from molass_legacy.AutorgKek.AppVersion import get_com_version_string as get_version_string

# from MachineTypes       import get_display_resolution, get_chassistype_name, get_monitors

# TODO: import wmi into exe
# width, height = get_display_resolution()
# print( 'display width, height=', width, height )

# chassistype_name = get_chassistype_name()
# print( chassistype_name )

# TODO: threading because it takes a significant amountt of time
# for monitor in get_monitors():
#    print( 'monitor=', monitor )
"""
if chassistype_name == 'Desktop':
    default_main_geometry = '1080x750'
else:
    default_main_geometry = '1080x700'
"""

SHOW_TEXT_DICT = {
    'range_type'                :   {   0:'Required Minimum Quality',
                                        1:'Required Minimum Concentration',
                                        2:'Boundary Points from Seq No.',
                                        3:'Intelligent Strategy',
                                        4:'Decomposed Elution Range',
                                        5:'Microfluidic Elution Range',
                                        },
    'zx_boundary_method'        :   {   'AUTO':'Automatic', 'FIXED':'',         'NO':'No boundary'}
    }


default_main_geometry = '1080x500'
# 永続的な記憶をオブジェクト化する（メモリに取り込む）。
setting_file    = 'serial_settings.py'
setting_info    = None
settings_       = None
v1_flag = None

def get_locals_for_settings():
    from numpy import array, nan, inf
    from molass_legacy.DataStructure.PeakInfo import PeakInfo
    from molass_legacy.DataStructure.RangeInfo import DecompEditorInfo, RangeEditorInfo
    from molass_legacy.DataStructure.AnalysisRangeInfo  import PairedRange
    # from Extrapolation.PreviewData import PreviewData, PreviewOptions
    from molass_legacy.Mapping.MappingParams import MappingParams
    from molass_legacy.Trimming.TrimmingInfo import TrimmingInfo
    # from molass_legacy.Decomposer.ModelEvaluator import ModelEvaluator
    # from molass_legacy.Decomposer.DualEvaluator import DualEvaluator
    # from molass_legacy.Models.ElutionCurveModels import EGHA, EMGA     # add required models such as STC, EDM, etc. when ready
    # from molass_legacy.Models.EGH import EGHA
    from molass_legacy._MOLASS.DummyClasses import (EGHA,
                                     EMGA,
                                     DualEvaluator,
                                     ModelEvaluator,
                                     # TrimmingInfo,
                                     # MappingParams,
                                     PreviewData, PreviewOptions,
                                     # PairedRange,
                                     # DecompEditorInfo, RangeEditorInfo,
                                     # PeakInfo,
                                     DummySd, DummyMapper,
                                     DummyJudgeHolder)  # wildcard import not allowed here
    return locals()

def initialize_settings(warn_on_fail=True):
    from molass_legacy.AutorgKek.Settings import overriding_init
    from molass_legacy.SerialAnalyzer.DevSettings import dev_init
    """
    must not be called in non-main processes
    """
    global setting_info, settings_
    setting_info    = PersistentPy(setting_file, locals_=get_locals_for_settings(), warn_on_fail=warn_on_fail)
    settings_       = setting_info.get_dictionary()
    check_version_stamp()
    overriding_init(settings_)
    dev_init(alt_folder=setting_info.alt_folder)

def get_settings_folder():
    return os.path.split( get_py_file_path(setting_file) )[0]

mask_           = None

stamp_name = 'version_stamp'

RANGE_TYPE_POINTS   = 2     # Boundary Points from molass_legacy.Elution No.
UV_BASE_NO          = 0
UV_BASE_CONST       = 1
UV_BASE_STANDARD    = 0
UV_BASE_SHIFTED     = 1
XRAY_BASE_NO        = 0
XRAY_BASE_CONST     = 1
XARY_BASE_LINEAR    = 1
XARY_BASE_QUADRATIC = 2
XARY_BASE_SPLINED   = 3
INTEGRAL_BASELINE = 5
DF_PATH_LENGTH_OLD  = 5.0
PF_PATH_LENGTH      = 8.437
SPRING8_PATH_LENGTH = 8.155
DF_PATH_LENGTH      = 7.071
DF_EXTINCTION       = 1

from molass.DataUtils.Beamline import BEAMLINE_NAME

def set_path_length(measurement_date, uv_device_no):
    if measurement_date is None or measurement_date > 20190930:
        beamline_name = BEAMLINE_NAME.get(uv_device_no, "")
        if beamline_name[0:2] == "PF":
            path_length = PF_PATH_LENGTH
        elif beamline_name[0:2] == "SP":
            path_length = SPRING8_PATH_LENGTH
        else:
            path_length = DF_PATH_LENGTH
    else:
        path_length = DF_PATH_LENGTH_OLD
    set_setting('path_length', path_length)

def get_beamline_name(uv_device_no=None, formal=False):
    beamline_name = get_setting("beamline_name")    # set in GuiMain
    if beamline_name == "":
        if uv_device_no is None:
            uv_device_no = get_setting("uv_device_no")  # set in SerialDataUtils.load_uv_file()
        beamline_name = BEAMLINE_NAME.get(uv_device_no, "")

    if formal:
        beamline_name = beamline_name.replace(" ", ", ").replace("PF", "Photon Factory")

    return beamline_name

ITEM_DEFAULTS = {
    stamp_name              : None,

    'in_folder'             : None,
    'uv_folder'             : None,
    'uv_file'               : None,
    'disable_uv_data'       : 0,
    'an_folder'             : None,
    'analysis_name'         : 'analysis-000',
    'analysis_folder'       : None,
    'fully_automatic'       : 1,
    'fully_optimize'        : 0,
    'auto_navigated_dailog' : 1,

    'data_exclusion'        : 0,
    'enable_usable_limit'   : True,

    'enable_mapping_anim'   : 0,
    'enable_drift_simulation'   : 0,
    'use_xray_conc'         : 0,
    'use_mtd_conc'          : 0,
    'mtd_file_path'         : 0,
    'apply_backsub'         : 0,

    'temp_folder'           : None,
    'auto_number'           : 1,
    'save_averaged_data'    : 1,
    'averaged_data_folder'  : None,
    'file_extension'        : '*.dat',
    'atsas_exe_paths'       : [],
    'num_recent_folders'    : 20,
    'recent_folders'        : {},

    'bad_mapping_ratios_ok' : 0,
    'bad_std_diff_ok'       : 1,
    'enable_auto_helper'    : 1,

    'use_moment_trimming'   : 0,

    'measurement_date'      : None,
    'sangler_version'       : None,
    'uv_restrict_list'      : None,
    'xr_restrict_list'      : None,
    'manually_trimmed'      : False,
    'uv_restrict_copy'      : None,
    'xr_restrict_copy'    : None,
    'cut_before_guinier'    : 2,
    'use_bqlimit'           : 1,
    'force_bqlimit'         : 0,
    'acceptable_rg_consist' : 0.8,

    'conc_adjust'           : 0,    # to be removed
    'conc_adjust_amount'    : None, # to be removed

    'manual_time_shift'     : None,
    'manual_time_scale'     : None,
    'mapper_sync_options'   : None,

    'uv_lpm_option'         : 1,
    'uv_baseline_opt'       : UV_BASE_CONST,
    'uv_baseline_type'      : 0,        # 0: no-correction, 1: linear, 4: shifted, 5: integral
    'uv_baseline_adjust'    : 1,
    'uv_baseline_with_bpa'  : 1,
    'uv_adjust_suppressed'  : 0,
    'uv_wl_lower_bound'     : 250,
    'enable_lrf_baseline'   : 1,
    'uv_device_no'          : "",
    'beamline_name'         : "",

    'baseline_manually'     : 0,        # 0: auto  1: manual
    'xray_baseline_type'    : None,     # 0: no-correction, 1: linear, 5: integral
    'xr_bpa_option'         : 1,
    'manual_end_slices'     : None,
    'xray_baseline_opt'     : XRAY_BASE_CONST,
    'xray_baseline_adjust'  : 0,
    'disable_xray_adjust'   : 1,
    'xray_baseline_with_bpa' : 1,
    'allow_angular_slope_in_mf' : 0,

    'enable_separate_fouling' : 0,
    'unified_baseline_type' : 1,        # baseline_type

    'base_drift_params'     : None,

    'dev_allow_ratio'       : 0.5,      # valid only only when adjust_both; will not be changed when disable_xray_adjust==1

    'fixed_guinier_start'   : 0,
    'guinier_start_point'   : 0,

    'baseline_auto'         : 1,
    'absorbance_baseline_type'  : 0,

    'scattering_base'       : 1,
    'baseline_degree'       : 1,
    'scattering_correction' : 0,
    'correction_iteration'  : 2,
    'peak_mapping_only'     : 0,
    'lpm_variations'        : 0,
    'mapping_show_mode'     : 'locally',
    'oopt_fit_consistency'  : 1,
    'oopt_optimize_only'    : 0,
    'oopt_qrg_limits'       : 'Limited',
    'oopt_qrg_limits_vals'  : [ 0.0, 1.3 ],
    'oopt_qrg_limits_apply' : 1,
    'min_quality'           : 0.5,
    'min_conc'              : 0.5,
    'range_type'            : 3,        # Intelligent Strategy
    'manual_range_info'     : None,
    'analysis_range_info'   : None,
    'preview_params'        : None,
    'num_ranges'            : 1,
    'input_smoothing'       : 1,        # this is forced to be 1 in SerialController.py
    'num_partitions_gpr'    : 200,
    'num_curves_averaged'   : 1,
    'avg_file_postfix'      : '_avg',
    'result_book'           : 'analysis_report.xlsx',
    'path_length'           : DF_PATH_LENGTH,
    'extinction'            : DF_EXTINCTION,
    'conc_factor'           : None,
    'absorbance_picking'    : 280,
    'absorbance_picking_sub'    : 260,
    'zero_absorbance'       : 400,      # can be overridden by flat_wavelength. see UV.PlainCurveUtils.get_flat_wavelength
    'zero_absorbance_auto'  : 1,
    'flat_wavelength'       : None,
    'consider_scatter'      : 0,
    'scatter_picking'       : None,
    'num_points_absorbance' : 1,
    'intensity_picking'     : 0.020,
    'num_points_intensity'  : 11,
    'x_ecurve_pickmethod'  : 0,
    'x_ecurve_pickslice'   : None,
    'x_ecurve_picking_q'   : None,
    'quality_weighting'     : [
                                0.2,    # basic_quality
                                0.2,    # positive_score
                                0.2,    # end_consistency
                                0.2,    # rg_stdev_score
                                0.2,    # q_rg_score
                              ],
    'suppress_low_quality_warning' : 0,
    'locally_2d_regression' : True,
    'using_elution_models'  : 0,
    'use_elution_models'    : 0,
    'has_elution_models'    : False,
    'matrix_formulation'    : 1,

    'enable_synthesized_lrf'    : 0,
    'synthesized_lrf'       : 0,
    'narrow_trimming'       : 1,

    'aq_smoothness'         : 0,
    'aq_positivity'         : 0,
    'penalty_weighting'     : [ 0.1, 0.01, 0.05, 0.005 ],
    'ignore_all_bqs'        : 0,
    'ignore_bq_list'        : None,
    'decomp_editor_info'    : None,
    'range_editor_info'     : None,
    'editor_ranges'         : None,
    'editor_model'          : None,
    'preview_model'         : None,
    'tau_hints_dict'        : None,
    'known_info_list'       : None,
    'svd_reconstruct'       : 1,
    'rank_control'          : 1,
    'allow_rank_variation'  : 1,
    'rank_increment'        : 0,
    'extended_conc_dep'     : 0,
    'conc_dependence'       : None,
    'enable_new_features'   : 0,
    'lrf_bound_correction'  : 0,
    'mapper_cd_color_info'  : None,
    'cd_eval_qmax'          : 0.2,
    'denat_dependent'       : 1,
    'enable_conc_opts'      : 1,
    'conc_curve_type'       : 0,    # 0: mapped UV, 1: scaled Xray
    'weight_matrix_type'    : 0,
    'concentration_datatype' : None,    # 0: XR model, 1: XR data, 2: UV model, 3: UV data
    'enable_conctype_change' : 0,

    'ee_elements_matrix'    : None,

    'zx_penalty_matrix'     : None,
    'zx_num_q_points'       : 3,
    'zx_build_method'       : 'MAX',
    'zx_boundary_method'    : 'AUTO',
    'zx_boundary'           : 0.1,
    'zx_a_file'             : 'zx_A_file.dat',
    'zx_b_file'             : 'zx_B_file.dat',
    'almerge_analyzer'      : 0,
    'axis_direction_desc'   : 0,
    'advance_persist'       : 0,
    'maintenance_mode'      : 0,
    'enable_debug_plot'     : 1,    # 1 also for release versions
    'enable_affine_tran'    : 1,
    'enable_gr_analyzer'    : 0,
    'enable_edm_model'      : 1,
    'decomp_from_separation'    : 0,

    'suppress_warning'      : False,
    'test_pattern'          : None,
    'mapping_image_folder'  : None,
    'decomp_image_folder_xray' : None,
    'decomp_image_folder_uv'   : None,
    'preview_image_folder'  : None,
    'no_excel_warning'      : True,

    'revoke_atsas'          : 0,
    'revoke_excel'          : 0,
    'revoke_cuda'           : 0,

    'auto_n_components'     : 1,
    'n_components'          : 10,
    'max_iterations'        : 200,
    'qmm_separately'        : 0,
    'fixed_random_seeds'    : None,
    'last_random_seeds'     : None,
    'forced_denoise_rank'   : None,
    'last_denoise_rank'     : None,
    'qmm_window_slice'      : None,
    'qmm_window_slice_uv'   : None,
    'denss_fitted_rg'       : 1,
    'found_lacking_q_values' : 0,
    'keep_tempfolder_averaged'  : 0,

    'used_mapping_params'   : None,
    'mapping_canvas_debug'  : 0,
    'beta_release'          : True,
    'using_shared_memory'   : True,

    'default_columtype_id'  : 'ad200w',
    'test_columtype_id'     : None,
    'exclusion_limit'       : 1300,     # Kda
    'poresize'              : None,
    'poresize_bounds'       : None,
    'num_plates_pc'         : 14400,    # per column (30cm)
    'columntype_id'         : None,

    'rg_curve_folder'       : None,
    'reuse_analysis_folder' : 0,
    'elution_model'         : 0,        # 0: Free EGH, 1 : Monopore, 2: LJ EGH, 3: FD EMG, 4: RT EMG, 5: EDM
    'init_sec_params'       : None,

    'optimizer_folder'      : None,     # optimized
    'optjob_folder'         : None,     # optimized/jobs, or optimized/jobs/nnn
    'optworking_folder'     : None,     # optimized/jobs/nnn;  should be unified with the above optjob_folder

    'rg_buffer_file'        : None,
    'default_func_egh'      : 'G0346',
    'default_func_sdm'      : 'G1100',
    'default_func_lj_egh'   : 'G0525',
    'default_func_fd_emg'   : 'G0665',
    'default_func_rt_egh'   : 'G0705',
    'default_func_edm'      : 'G2010',
    'default_objective_func' : None,
    'TAU_BOUND_RATIO'       : 0.65,
    'RG_UPPER_BOUND'        : 150,
    'RATE_R_UPPER_BOUND'    : 300,  # > 228 for 20200623_1
    'recompute_rg_curve'    : 0,

    'ratio_interpretation'  : 0,
    'separate_eoii'         : 0,
    'separate_eoii_type'    : 0,
    'separate_eoii_flags'   : [],
    'apply_rg_discreteness' : 0,
    'rg_discreteness_unit'  : 1.0,
    'apply_mw_integrity'    : 0,
    'mw_integer_ratios'     : None,
    'avoid_peak_fronting'   : 0,
    'kratky_smoothness'     : 1,
    'ignore_secconformance' : 0,
    'try_model_composing'   : 0,
    'identification_allowance' : 0.1,
    'apply_sf_bounds'       : 1,
    'sf_bound_ratio'        : 1.0,
    'NUM_MAJOR_SCORES'      : 7,
    'trust_rg_curve_folder' : False,    # temporary fix to trimming inconsistency

    # See also Optimizer.OptimizerSettings.py on the items below
    't0_upper_bound'        : None,
    'uv_basemodel'          : 1,
    'poreexponent'          : None,

    'local_debug'           : False,
    'debug_path'            : None,
    'debug_fh'              : None,

    'report_default_font'   : "Arial",
    'suppress_numba_warning' : True,
   }

V2_TEMPORARY_ITEMS = [
    'uv_wl_lower_bound',    # to be refreshed always without clearing the permanent memory
    'using_shared_memory', 'reuse_analysis_folder', 'elution_model', 'unified_baseline_type', 'init_sec_params',
    'optimizer_folder', 'optjob_folder', 'optworking_folder',
    'rg_buffer_file', 'uv_basemodel', 'default_objective_func',
    'RATE_R_UPPER_BOUND',
    't0_upper_bound', 'poreexponent',
    'poresize', 'exclusion_limit', 'poresize_bounds',
    'separate_eoii', 'separate_eoii_type', 'separate_eoii_flags',
    'apply_rg_discreteness', 'rg_discreteness_unit',
    'apply_mw_integrity', 'mw_integer_ratios',
    'avoid_peak_fronting',
    'ignore_secconformance', 'try_model_composing', 'identification_allowance', 'apply_sf_bounds', 'sf_bound_ratio',
    'NUM_MAJOR_SCORES',
    'trust_rg_curve_folder',
    ]

TEMPORARY_ITEMS = [
    'revoke_atsas', 'revoke_excel', 'revoke_cuda',

    'uv_folder', 'uv_file', 'disable_uv_data',
    'averaged_data_folder',

    'fully_optimize',

    'bad_mapping_ratios_ok', 'bad_std_diff_ok',

    'fixed_guinier_start', 'guinier_start_point',

    'manual_time_shift', 'manual_time_scale', 'mapper_sync_options',
    'conc_adjust', 'conc_adjust_amount',
    'baseline_auto', 'absorbance_baseline_type',
    'scattering_base', 'baseline_degree', 'scattering_correction',
    'correction_iteration',
    'consider_scatter',
    'scatter_picking',
    'flat_wavelength',
    'x_ecurve_pickmethod', 'x_ecurve_pickslice', 'x_ecurve_picking_q',

    'enable_usable_limit',
    'use_xray_conc', 'use_mtd_conc',
    'measurement_date',
    'uv_wl_lower_bound',    # to be refreshed always without clearing the permanent memory
    'uv_restrict_list', 'xr_restrict_list', 'manually_trimmed',
    'uv_restrict_copy', 'xr_restrict_copy',

    'uv_lpm_option',
    'uv_baseline_opt', 'uv_baseline_type', 'uv_baseline_adjust', 'uv_adjust_suppressed',
    'uv_baseline_with_bpa', 'uv_device_no', 'beamline_name',
    'baseline_manually', 'xr_bpa_option', 'manual_end_slices',
    'xray_baseline_opt', 'xray_baseline_type', 'xray_baseline_adjust',
    'xray_baseline_with_bpa',
    'apply_backsub',
    'dev_allow_ratio',
    'base_drift_params',

    'conc_factor',
    'conc_adjust', 'conc_adjust_amount',

    'analysis_range_info', 'preview_params',
    'range_type', 'manual_range_info', 'num_ranges',

    'path_length', 'extinction',
    'zx_penalty_matrix',
    'ignore_all_bqs', 'ignore_bq_list',
    'decomp_editor_info', 'range_editor_info',
    'editor_ranges', 'editor_model', 'preview_model', 'tau_hints_dict',
    'known_info_list',
    'svd_reconstruct', 'rank_increment', 'conc_dependence', 'weight_matrix_type',
    'lrf_bound_correction', 'mapper_cd_color_info',
    'using_elution_models',
    'use_elution_models',
    'has_elution_models',
    'concentration_datatype',

    'columntype_id', 'test_columtype_id',

    'auto_n_components', 'n_components', 'max_iterations', 'qmm_separately',
    'fixed_random_seeds', 'last_random_seeds', 'forced_denoise_rank', 'last_denoise_rank',
    'qmm_window_slice', 'qmm_window_slice_uv', 'denss_fitted_rg',
    'found_lacking_q_values',
    'used_mapping_params', 'mapping_canvas_debug',

    'suppress_low_quality_warning',
    'rg_curve_folder',
    'local_debug',
    'debug_path',
    'debug_fh',
    *V2_TEMPORARY_ITEMS,
    ]

XRAY_CONC_ITEM_DEFAULTS = {
    'uv_baseline_opt'       : 0,
    'uv_baseline_type'      : 0,
    'uv_baseline_adjust'    : 0,
    'xray_baseline_opt'     : 1,
    'xray_baseline_type'    : 1,
    'xray_baseline_adjust'  : 0,
    }

MICROFLUIDIC_ITEM_DEFAULTS = {
    'uv_baseline_opt'       : 0,
    'uv_baseline_type'      : 0,
    'uv_baseline_adjust'    : 0,
    'xray_baseline_opt'     : 0,
    'xray_baseline_type'    : 0,
    'xray_baseline_adjust'  : 0,
    }

NO_SAVE_ITEMS = [
    'test_pattern',
    'mapping_image_folder',
    'decomp_image_folder_xray',
    'decomp_image_folder_uv',
    'preview_image_folder',
    ]

ALTERNATIVE_WEIGHTS = [ 0.2, 0.2, 0.3, 0.2, 0.1 ]

def reload_settings():
    global setting_info
    global settings_

    setting_info    = PersistentPy(setting_file, locals_=get_locals_for_settings())
    settings_       = setting_info.get_dictionary()

def clear_settings():
    global settings_
    settings_ = {}
    if setting_info is None:
        initialize_settings()
    setting_info.set_dictionary( settings_ )

def get_setting( item ):
    assert( item in ITEM_DEFAULTS )
    if settings_ is None:
        initialize_settings()

    value = settings_.get( item )
    if value is None:
        value = ITEM_DEFAULTS.get( item )
        set_setting( item, value )
    return value

def get_xray_picking():
    picking = get_setting('x_ecurve_picking_q')
    if picking is None:
        picking = get_setting('intensity_picking')
    return picking

def restore_default_setting( item ):
    set_setting( item, ITEM_DEFAULTS.get( item ) )

def set_setting( item, value ):
    assert( item in ITEM_DEFAULTS )
    if settings_ is None:
        initialize_settings()

    settings_[ item ] = value

def reset_setting( item ):
    assert( item in ITEM_DEFAULTS )
    settings_[item] = ITEM_DEFAULTS[item]

def temporary_settings_begin():
    global settings_save
    settings_save = settings_.copy()

def temporary_settings_end():
    global settings_save
    global settings_
    settings_ = settings_save
    setting_info.set_dictionary( settings_ )

def save_settings(file=None, clear_no_save_items=False, debug=False):
    if clear_no_save_items:
        for item in NO_SAVE_ITEMS:
            set_setting( item, ITEM_DEFAULTS.get( item ) )

    if debug:
        settings_ = setting_info.get_dictionary()
        with open(r"D:\TODO\20230724\temp\setting-debug.txt", "w") as fh:
            for k, v in settings_.items():
                fh.write(str((k, v)) + "\n")

    setting_info.save( file=file )

def load_settings_dict(file):
    from molass_legacy.KekLib.PersistentPy import load_py_file
    dict_ = load_py_file(file, locals_=get_locals_for_settings())
    # TODO: compatibillty check
    return dict_

def load_settings(file):
    initialize_settings()
    dict_ = load_settings_dict(file)
    setting_info.set_dictionary(dict_)

def clear_temporary_settings():
    for item in TEMPORARY_ITEMS:
        set_setting( item, ITEM_DEFAULTS[item] )

def clear_v2_temporary_settings():
    for item in V2_TEMPORARY_ITEMS:
        set_setting( item, ITEM_DEFAULTS[item] )

def do_xray_conc_temporary_settings():
    # care for the fact that defaults are not the same for microfluidic processing
    for k, v in XRAY_CONC_ITEM_DEFAULTS.items():
        set_setting(k, v)

def do_microfluidic_temporary_settings():
    # care for the fact that defaults are not the same for microfluidic processing
    for k, v in MICROFLUIDIC_ITEM_DEFAULTS.items():
        set_setting(k, v)

def check_version_stamp():
    version_stamp = get_setting( stamp_name )
    current_version = get_version_string()
    if current_version.find("_MOLASS 1.0") >= 0:
        modify_to_v1_0_setting()
    if version_stamp is None or version_stamp != current_version:
        # TODO: MessageBox
        clear_settings()
    set_setting( stamp_name, get_version_string() )
    # dump_settings()

def modify_to_v1_0_setting():
    ITEM_DEFAULTS['enable_synthesized_lrf'] = 0
    ITEM_DEFAULTS['synthesized_lrf'] = 0
    ITEM_DEFAULTS['narrow_trimming'] = 1

def dump_settings():
    fh = open("settings.txt", "w")
    sorted_list = sorted(settings_.items(), key=lambda x: x[0])
    fh.write('\n'.join([str(pair) for pair in sorted_list]))
    fh.close()



"""
    Optimizer.Strategies.EghSheet.py

    Copyright (c) 2024, SAXS Team, KEK-PF
"""

def get_data_list(num_components):
    data_list = []
    for i in range(num_components):
        data_list.append(("h_%02d"%i, "", "1", ""))
        data_list.append(("mu_%02d"%i, "", "1", ""))
        data_list.append(("sigma_%02d"%i, "", "1", ""))
        data_list.append(("tau_%02d"%i, "", "1", ""))

    for i in range(num_components):
        data_list.append(("rg_%02d"%i, "", "", "1"))
    
    data_list.append(("xr_base_a", "1", "", ""))
    data_list.append(("xr_base_b", "1", "", ""))

    for i in range(num_components):
        data_list.append(("uvh_%02d"%i, "", "1", ""))

    data_list.append(("uv_base_L", "1", "", ""))
    data_list.append(("uv_base_x0", "1", "", ""))
    data_list.append(("uv_base_k", "1", "", ""))
    data_list.append(("uv_base_b", "1", "", ""))
    data_list.append(("uv_base_s1", "1", "", ""))
    data_list.append(("uv_base_s2", "1", "", ""))
    data_list.append(("uv_base_diff", "1", "", ""))

    data_list.append(("mapping_a", "", "1", ""))
    data_list.append(("mapping_b", "", "1", ""))

    data_list.append(("mappable_a", "", "1", ""))
    data_list.append(("mappable_b", "", "1", ""))

    data_list.append(("Npc", "", "", "1"))
    data_list.append(("poresize", "", "", "1"))
    data_list.append(("tI", "", "", "1"))
    data_list.append(("t0", "", "", "1"))
    data_list.append(("P", "", "", "1"))
    data_list.append(("m", "", "", "1"))
    return data_list


# coding: utf-8
"""
    SimpleGuinierGuiUtils.py

    Copyright (c) 2017, Masatsuyo Takahashi, KEK-PF
"""
import numpy        as np
from molass_legacy.KekLib.OurTkinter     import Tk, Font
from molass_legacy.KekLib.TkUtils        import adjusted_geometry

window_title    = 'Guinier Analyzer 0.0.1'

class CheckbuttonFrame( Tk.Frame ):
    def __init__( self, parent, types=None ):
        Tk.Frame.__init__( self, parent )

        if types is None:
            types = range(17)

        self.types = types
        self.clear_btn   = Tk.Button( self, text='clear all', command=self.cb_clear )
        self.clear_btn.grid( row=0, column=0 )

        self.cb_vars = []
        for i in types:
            cb_var = Tk.IntVar()
            cb_var.set( 1 )
            self.cb_vars.append( cb_var )
            cb = Tk.Checkbutton( self, variable=cb_var, text='type %d' % i )
            cb.grid( row=i+1, column=0, sticky=Tk.W )
            cb_var.trace( 'w', self.cb_vars_tracer )

        self.cb_vars_tracer_reset()

    def cb_vars_tracer_reset( self ):
        self.changed = False

    def cb_vars_tracer( self, *args ):
        self.changed = True

    def cb_clear( self ):
        text = self.clear_btn.cget( 'text' )
        val = 0 if text == 'clear all' else 1
        for cb_var in self.cb_vars:
            cb_var.set( val )
        text = 'check all' if val == 0 else 'clear all'
        self.clear_btn.config( text=text )

    def make_index( self, z ):
        types = []
        for t, v in enumerate( [ var.get() for var in self.cb_vars ] ):
            if v  == 1:
                types.append( self.types[t] )

        self.cb_vars_tracer_reset()

        if len( types ) == len( self.cb_vars ):
            return slice(None, None)
        else:
            return np.logical_or.reduce( [ z == t for t in types ] )

class TypeInfoPanel(Tk.Toplevel):
    def __init__( self, parent ):
        self.parent = parent
        Tk.Toplevel.__init__( self, parent )
        self.title( 'Type Help'  )
        self.fixed_font = Font.Font( family="Courier", size=9 )

    def show( self, type_names ):
        frame = Tk.Frame( self )
        frame.pack( padx=50, pady=20 )
        self.type_names = type_names
        for i, name in enumerate( self.type_names ):
            label = Tk.Label( frame, text='type %2d:    %s' % ( i, name ), font=self.fixed_font )
            label.grid( row=i, column=0, sticky=Tk.W )
        self.update()
        self.geometry( adjusted_geometry(self.geometry()) )

class FolderInfoPanel(Tk.Toplevel):
    def __init__( self, parent ):
        self.parent = parent
        Tk.Toplevel.__init__( self, parent )
        self.title( 'Folder Help'  )
        self.fixed_font = Font.Font( family="Courier", size=9 )

    def show( self, folder_paths ):
        frame = Tk.Frame( self )
        frame.pack( padx=20, pady=20 )
        self.folder_paths = folder_paths
        for i, path in enumerate( self.folder_paths ):
            button = Tk.Button( frame, text='folder %03d: %s' % ( i, path ), font=self.fixed_font,
                        command=lambda i_=i: self.draw_folder(i_),
                        relief=Tk.FLAT )
            j = i % 20
            k = i //20
            button.grid( row=j, column=k, sticky=Tk.W, padx=20 )
        self.update()
        self.geometry( adjusted_geometry(self.geometry()) )

    def draw_folder( self, i ):
        self.parent.draw_folder( folder_no=i )



# coding: utf-8
"""

    OpenPyXlUtils.py

    Copyright (c) 2019, Masatsuyo Takahashi, KEK-PF

"""
from openpyxl import Workbook, load_workbook

def merge_into_a_book( input_books, merged_book, progress_cb=None ):
    """
        it is not supported in openpyxl.
        instead, you can do as follows.

        wb = Workbook()
        wb.remove(wb.active)

        for i in range(3):
            ii = i + 1
            ws = wb.create_sheet('TestSheet%d' % ii)
            # edit ws

        wb.save(out_book)

        See the current state in the following pages.

        copy a worksheet in openpyxl
        https://stackoverflow.com/questions/34808394/copy-a-worksheet-in-openpyxl

        "I'm afraid copying worksheets is not supported because it is far from easy to do."

        Add a new sheet to a existing workbook in python
        https://stackoverflow.com/questions/40385689/add-a-new-sheet-to-a-existing-workbook-in-python
    """
    pass



# coding: utf-8
"""

    InputOutputDisplay.py

    Copyright (c) 2017-2018, Masatsuyo Takahashi, KEK-PF

"""

from molass_legacy.KekLib.OurTkinter             import Tk, Dialog, is_empty_val
from molass_legacy._MOLASS.SerialSettings         import get_setting

class InputOutputDisplay( Tk.Frame ):
    def __init__( self, parent, scatter_folder=None, absorb_file=None, book_path=None, label_witdh=100, value_width=300 ):
        Tk.Frame.__init__( self, parent )

        if scatter_folder is None:
            scatter_folder  = get_setting( 'in_folder' )

        if absorb_file is None:
            absorb_file     = '/'.join( [ get_setting( 'uv_folder' ), get_setting( 'uv_file' ) ] )

        if book_path is None:
            book_path       = '/'.join( [ get_setting( 'analysis_folder' ), get_setting( 'result_book' ) ] )

        label_width_frame = Tk.Frame( self, width=label_witdh )
        label_width_frame.grid( row=0, column=0 )
        value_width_frame = Tk.Frame( self, width=value_width )
        value_width_frame.grid( row=0, column=1 )

        scatter_label   = Tk.Label( self, text="Xray Scattering Data" )
        scatter_label.grid( row=0, column=0, sticky=Tk.E, pady=1 )

        scatter_value   = Tk.Label( self, text=scatter_folder, bg='white', anchor=Tk.W )
        scatter_value.grid( row=0, column=1, sticky=Tk.W + Tk.E, padx=5 )

        if get_setting('use_xray_conc') == 0:
            uv_state    = Tk.NORMAL
            uv_bg       = 'white'
        else:
            uv_state    = Tk.DISABLED
            uv_bg       = None
            absorb_file = "Using Xray-proportional concentration"

        absorb_label    = Tk.Label( self, text="UV Absorbance Data", state=uv_state )
        absorb_label.grid( row=1, column=0, sticky=Tk.E, pady=1 )

        absorb_value    = Tk.Label( self, text=absorb_file, bg=uv_bg, anchor=Tk.W, state=uv_state )
        absorb_value.grid( row=1, column=1, sticky=Tk.W + Tk.E, padx=5 )

        outbook_label   = Tk.Label( self, text="Result Book" )
        outbook_label.grid( row=2, column=0, sticky=Tk.E, pady=1 )

        outbook_value    = Tk.Label( self, text=book_path, bg='white', anchor=Tk.W )
        outbook_value.grid( row=2, column=1, sticky=Tk.W + Tk.E, padx=5 )



"""
    DebugImpl.py

    Copyright (c) 2023-2024, SAXS Team, KEK-PF
"""
import molass_legacy.KekLib.DebugPlot as plt
from molass_legacy.Elution.CurveUtils import simple_plot

def debug_impl_ConcDepend(self):
    print("debug_impl")
    rdr_hints, ecurve, judge_holder = self.debug_info
    major_peak_info = ecurve.get_major_peak_info()
    print(len(rdr_hints), len(ecurve.peak_info), len(major_peak_info))

    x = ecurve.x
    y = ecurve.y

    with plt.Dp():
        fig, ax = plt.subplots()
        # simple_plot(ax, ecurve)
        ax.plot(x, y)
        for rec in ecurve.get_major_peak_info():
            px = rec[1]
            ax.plot(px, ecurve.spline(px), "o", color="red")
        fig.tight_layout()
        plt.show()


def debug_impl_SummaryBook(self):
    from openpyxl import Workbook
    from importlib import reload
    import Reports.SummaryBook
    reload(Reports.SummaryBook)
    from molass_legacy.Reports.SummaryBook import SummaryBook

    book_file, controller = self.debug_info
    print("debug_impl: book_file=", book_file)

    wb = Workbook()
    summary = SummaryBook(wb, controller)
    wb.save(book_file)

def debug_impl(self):
    print("debug_impl")
    with plt.Dp():
        fig, axes = plt.subplots(ncols=5, figsize=(15, 3))
        for ax in axes:
            ax.plot(0, 0, "o")
        fig.tight_layout()
        plt.show()


# coding: utf-8
"""
    ConvolutionDemo.py

    Copyright (c) 2018, Masatsuyo Takahashi, KEK-PF
"""
import numpy as np
from scipy.integrate import quad
import matplotlib.pyplot as plt
import matplotlib.gridspec  as gridspec

NEGATIVE_INF    = -20
POSITIVE_INF    =  20

class Convolution:
    def __init__( self, f, g, lower_limit=NEGATIVE_INF, upper_limit=POSITIVE_INF, zero_neg=False ):
        self.f  = f
        self.g  = g
        self.lower_limit = lower_limit
        self.upper_limit = upper_limit
        self.zero_neg = zero_neg

    def eval_scaler( self, t ):
        h = lambda x: self.f( x ) * self.g( t - x )
        if self.zero_neg:
            return quad( h, self.lower_limit, t )[0]
        else:
            return quad( h, self.lower_limit, self.upper_limit )[0]

    def __call__( self, t ):
        if np.isscalar( t ):
            return self.eval_scaler( t )
        else:
            return np.array( [ self.eval_scaler(x) for x in t ] )

class Gaussian:
    def __init__( self, mu, sigma ):
        self.mu     = mu
        self.sigma  = sigma
        self.c      = 1 / np.sqrt( 2 * np.pi * sigma**2 )

    def __call__( self, x ):
        return self.c * np.exp( - (x - self.mu)**2/( 2*self.sigma**2 ) )

class Exponential:
    def __init__( self, k):
        self.k  = k

    def __call__( self, x ):
        if np.isscalar( x ):
            return 0 if x < 0 else self.k * np.exp( -self.k * x )
        else:
            x_pos   = x >= 0
            v_pos   = self.k * np.exp( -self.k * x[x_pos] )
            return np.hstack( [ np.zeros( len(x) - len(v_pos) ), v_pos ] )

class EmgConvolutionDemo:
    def __init__( self ):
        k   = 1
        m   = 3
        s   = 0.5
        n   = Gaussian( m, s )
        e   = Exponential( k )
        h   = Convolution( n, e, zero_neg=True )

        x   = np.linspace( -3, 10, 1000 )

        fig = plt.figure( figsize=(12, 9) )

        ax1 = fig.add_subplot( 111 )

        ax1.set_title( "EMG - Convolution of Gaussian and Exponential" )

        ax1.plot( x, n(x), label='G(%g, %g)' % (m, s) )
        ax1.plot( x, e(x), label='E(%g)' % (k) )
        ax1.plot( x, h(x), label='EMG - Convolution of G(%g, %g) and E(%g)' % (m, s, k) )

        ax1.legend()
        fig.tight_layout()
        plt.show()



# coding: utf-8
"""
    MethodFile.py

    Copyright (c) 2018-2019, SAXS Team, KEK-PF
"""
from bisect import bisect_left
import numpy as np

START_DELAY_RATIO = 0.1
TOTAL_RATE_COLUMNS = [2, 4, 6, 10]

class MethodFile:
    def __init__(self, filepath):
        self.filepath = filepath
        self.rows = []
        self.rows_integ = []
        self.x = None
        self.y = None
        self.computed_size = None

        t = 0
        with open(filepath) as fh:
            is_in_data_section  = False
            for line in fh:
                # print(line)
                if line.find('[Data]') >= 0:
                    is_in_data_section = True
                    continue
                if line.find('[End]') >= 0:
                    is_in_data_section = False
                    continue
                if is_in_data_section:
                    if line[0] == ':':
                        self.colnames = line[1:].split('\t')
                        continue
                    values = [ eval(value) for value in line.split('\t') ]
                    self.rows.append(values)
                    m = values[1]/60
                    t += m
                    total_rate = np.sum([values[k] for k in TOTAL_RATE_COLUMNS])
                    self.rows_integ.append([m, t, total_rate])

    def get_elution_data(self, size, debug=False, auto=True):
        print('get_elution_data', size)
        if self.computed_size is None or size != self.computed_size:
            x, y = self.compute_elution_data(size, debug=debug, auto=auto)
            self.x, self.y = x, y
            self.computed_size = size

        return self.x, self.y

    def compute_elution_data(self, size, tv_only=False, debug=False, auto=True):
        print('compute_elution_data', size, tv_only)
        tlist = []
        t = 0
        values = []
        check_rownum = True
        self.row_nums = []
        for k, row in enumerate(self.rows):
            number = row[0]
            # assert number == k + 1
            if check_rownum:
                if number == k + 1:
                    pass
                else:
                    import OurMessageBox as MessageBox
                    MessageBox.showwarning( "Warning",
                        "Row numbers are not as expected in %s." % self.filepath,
                        parent=None )
                    check_rownum = False
            duration = row[1]
            tlist.append(t)
            t += duration
            tlist.append(t)
            fr_values = []
            to_values = []
            for j in range(5):
                j_ = 2+j*2
                fr_value = row[j_]
                fr_values.append(fr_value)
                to_value = row[j_+1]
                to_values.append(to_value)
            values.append(fr_values)
            values.append(to_values)

        self.t = t = np.array(tlist)
        matrix = np.array(values)
        print('matrix.shape=', matrix.shape, 'len(t)=', len(t))
        total = np.sum(matrix[:,[0,1,2,4]], axis=1)
        # print('t=', t)
        # print('total=', total)
        j = 1
        half_t = t[-1] * 0.7
        print('half_t=', half_t)
        self.v = v = matrix[:,j]/total
        self.start = None
        for ub in [0.2]:
            wh = np.where( np.logical_and( t < half_t, v < ub ) )[0]
            if len(wh) > 0:
                self.start = wh[-1]
                break

        assert self.start is not None

        try:
            self.start_t = t[self.start]
            self.end = np.where( np.logical_and(t > self.start_t, v > 0.5) )[0][0]
            print( 'start, end=', (self.start, self.end) )
            self.end_t = t[self.end]
            self.start_v = v[self.start]
            self.end_v = v[self.end]
        except:
            import molass_legacy.KekLib.DebugPlot as plt
            fig = plt.figure()
            ax = fig.gca()
            ax.plot(t, v)
            plt.show()

        if tv_only:
            return t, v

        x = np.linspace(0, t[-1], size)
        y = []
        for x_ in x:
            i = bisect_left(t, x_)
            if self.start_t <= x_ and x_ <= self.end_t:
                w = (x_ - self.start_t)/(self.end_t - self.start_t)
                v_ = self.start_v * (1-w) + self.end_v * w
            else:
                v_ = v[i]
            y.append(v_)

        if debug:
            import molass_legacy.KekLib.DebugPlot as plt
            import time
            # from molass_legacy.KekLib.NumpyUtils import np_savetxt
            # np_savetxt('matrix.csv', matrix)
            colname = self.colnames[2+j*2].replace('St.', '')
            fig = plt.figure()
            ax = fig.gca()
            # for j in range(matrix.shape[1]):
            ax.plot(x, y, label=colname + ' generated')
            ax.plot(t, v, 'o', label=colname + ' data')
            for k in [self.start, self.end]:
                ax.plot(t[k], v[k], 'o', color='red')
            ax.legend()
            fig.tight_layout()
            if auto:
                plt.show(block=False)
                time.sleep(0.5)
            else:
                plt.show()

        self.cp = []
        for t_ in [self.start_t, self.end_t]:
            self.cp.append(bisect_left(x, t_))
        print('self.cp=', self.cp)
        return x, np.array(y)

    def get_linear_slope_ends(self):
        # print('cp=', self.cp)
        return [ self.cp[0], self.cp[1]-1 ]

    def guess_range(self, size):
        self.get_elution_data(size)
        start, stop = self.get_linear_slope_ends()
        length = stop - start
        delay = int(length*START_DELAY_RATIO)
        return slice(start+delay, stop)

    def guess_critical_points(self):
        last_row_integ = self.rows_integ[-1]
        print(last_row_integ)
        size = int(last_row_integ[1])+1
        t, v = self.compute_elution_data(size, tv_only=True)
        w = np.where(np.logical_and(t < self.start_t, v > self.start_v))
        i = w[0][-1]

        if False:
            import molass_legacy.KekLib.DebugPlot as plt
            n = np.arange(len(t))
            print(list(zip(n, t, v)))
            print([i, self.start, self.end])
            plt.plot(t, v, ':')
            plt.plot(t, v, 'o')
            for k in [i, self.start, self.end]:
                plt.plot(t[k], v[k], 'o', color='red')
            plt.show()

        i_ = i//2
        j_ = self.start//2
        k_ = j_ + 1
        return i_, j_, k_



"""
    DiniSigmoid.py

    from:

        Normalized tunable sigmoid functions
        https://dinodini.wordpress.com/2010/04/05/normalized-tunable-sigmoid-functions/

        Normalized Tunable Sigmoid Function
        https://dhemery.github.io/DHE-Modules/technical/sigmoid/#function

"""
import numpy as np
from pynverse import inversefunc

def dini_sigmoid(x, k):
    return x*(1 - k)/(k - 2*k*np.abs(x) + 1)

def dini_sigmoid_inv(y, k):
    def func(y):
        return dini_sigmoid(y, k)
    return inversefunc(func, y_values=y)

def demo():
    import matplotlib.pyplot as plt

    fig, ax = plt.subplots()
    x = np.linspace(-1, 1, 100)
    for k in np.linspace(-0.95, 0.95, 9):
        y = dini_sigmoid(x, k)
        ax.plot(x, y, label="k=%.3g" % k)

    k = -0.95
    y_ = 0.9
    x_ = dini_sigmoid_inv(y_, k)
    ax.plot(x_, y_, "o", color="red")

    ax.legend()
    plt.show()

if __name__ == '__main__':
    demo()



# coding: utf-8
"""
    DpiAware.py

    Copyright (c) 2020, SAXS Team, KEK-PF
"""

def set_dpi_aware():
    """
    setting suitable layout for high-resolution monitors

    https://github.com/asweigart/pyautogui/issues/33

    see also:
    Win32api is not giving the correct coordinates with GetCursorPos() in python
    https://stackoverflow.com/questions/32541475/win32api-is-not-giving-the-correct-coordinates-with-getcursorpos-in-python
    """
    from ctypes import windll
    user32 = windll.user32
    user32.SetProcessDPIAware()



"""
    Models.Stochastic.MonoporeUtilsImpl.py

    Copyright (c) 2024, SAXS Team, KEK-PF
"""
import numpy as np
import molass_legacy.KekLib.DebugPlot as plt
from SecTheory.BasicModels import robust_single_pore_pdf

def compute_monopore_curves(x, mnp_colparams, rgs, scales, return_moments=False):
    N, T, x0, me, mp, poresize = mnp_colparams
    rhov = rgs/poresize
    rhov[rhov > 1] = 1
    cy_list = []
    moments = []
    for rho, scale in zip(rhov, scales):
        ni_ = N * (1 - rho)**me
        ti_ = T * (1 - rho)**mp
        cy = scale * robust_single_pore_pdf(x - x0, ni_, ti_)
        cy_list.append(cy)
        M1 = x0 + N * T * (1 - rho)**(me + mp)
        M2 = 2 * N * T**2 * (1 - rho)**(me + 2*mp)
        moments.append((M1, M2))
    ty = np.sum(cy_list, axis=0)
    if return_moments:
        return cy_list, ty, moments
    else:
        return cy_list, ty

def plot_monopore_component_with_sliders_impl(title, x, y, peak_rgs, mnp_params, x0_upper, cy_list, ty, moments, egh_moments_list):
    from matplotlib.widgets import Slider, Button
    from molass_legacy.SerialAnalyzer.DataUtils import get_in_folder

    print("moments=", moments)

    plot_params = np.asarray(mnp_params).copy()
    print("plot_params=", plot_params)

    slider_specs = [    ("N", 0, 6000, mnp_params[0]),
                        ("T", 0, 3, mnp_params[1]),
                        ("t0",  -1000, 1000, mnp_params[2]),
                        ("me", 0, 3, mnp_params[3]),
                        ("mp", 0, 3, mnp_params[4]),
                        ("poresize", 20, 400, mnp_params[5]),
                        ]

    num_peaks = len(peak_rgs)
    with plt.Dp():
        fig, (ax1,ax2) = plt.subplots(ncols=2, figsize=(18, 5))
        fig.suptitle(title)
        ax2.set_title("Moments Fitting for %s" % get_in_folder())
        ax2.set_xlabel("Time (frames)")
        ax2.set_ylabel("Intensity")
    
        mlines = []
        ecurves = []
        for i, ax in enumerate((ax1,ax2)):
            ax.plot(x, y, color='orange', label='data')
            curve, = ax.plot(x, ty, ':', color="red", label='model total')
            if i == 1:
                ecurves.append(curve)
            for k, (cy, M, eghM) in enumerate(zip(cy_list, moments, egh_moments_list)):
                curve, = ax.plot(x, cy, ':', label='component-%d' % k)
                if i == 1:
                    ecurves.append(curve)
                label = None if k < num_peaks - 1 else "Monopore Moments"
                line = ax.axvline(x=M[0], color="green", label=label)
                if i == 1:
                    mlines.append(line)
                label = None if k < num_peaks - 1 else "EGH Moments"
                ax.axvline(x=eghM[0], ls=":", lw=3, color="cyan", label=label)
            ax.axvline(x=x0_upper, color="pink")

        def slider_update(k, val):
            print([k], "slider_update", val)
            plot_params[k] = val
            cy_list_, ty_, moments_ = compute_monopore_curves(x, plot_params[0:6], peak_rgs, plot_params[6:], return_moments=True)
            for cy, ecurve in zip([ty_] + cy_list_, ecurves):
                ecurve.set_ydata(cy)
            print("moments_=", moments_)
            for line, M in zip(mlines, moments_):
                line.set_xdata(M[0])
            fig.canvas.draw_idle()

        slider_axes = []
        sliders = []
        for k, (label, valmin, valmax, valinit) in enumerate(slider_specs, start=0):
            ax_ = fig.add_axes([0.75, 0.8 - 0.08*k, 0.2, 0.03])
            slider  = Slider(ax_, label=label, valmin=valmin, valmax=valmax, valinit=valinit)
            slider.on_changed(lambda val, k_=k: slider_update(k_, val))
            slider_axes.append(ax_)
            sliders.append(slider)

        def reset(event):
            print("reset")
            for k, slider in enumerate(sliders):
                slider.reset()

        button_ax = fig.add_axes([0.85, 0.2, 0.12, 0.05])
        debug_btn = Button(button_ax, 'Reset', hovercolor='0.975')
        debug_btn.on_clicked(reset)

        # xmin, xmax = ax1.get_xlim()
        # ymin, ymax = ax1.get_ylim()
        m1, m2 = moments[0]
        xmin = m1 - 10*np.sqrt(m2)
        m1, m2 = moments[-1]
        xmax = m1 + 10*np.sqrt(m2)                
        ax2.set_xlim(xmin, xmax)
        ax2.legend()
        fig.tight_layout()
        fig.subplots_adjust(right=0.7)
        ret = plt.show()

    return ret


"""
    Distance.FrobeniusXdiffmax.py

    Copyright (c) 2024, SAXS Team, KEK-PF
"""
import numpy as np

def frobenius_xdiffmax(y1, y2, adjust=0, max_mask=None, debug=False):
    diff = y1 - y2
    d1 = np.linalg.norm(diff) 
    if max_mask is None:
        diff_ = diff
    else:
        diff_ = diff[max_mask]  # to avoid over-esiamtion from outliers
    std = np.std(diff_)
    d2 = std + np.max(np.abs(diff_))/std
    if debug:
        print("d1=", d1, "d2=", d2)
    return np.log10(d1*d2) + adjust


"""
    Preprocess.py

    Copyright (c) 2022,2025, SAXS Team, KEK-PF
"""
import numpy as np

NEGATIVE_RATIO_LIMIT = 0.7  # < 0.85 for 20220716/lowSN_BSA_01

def correct_flat_negative_baseline(y, debug=False):

    negetive_ratio = len(np.where(y < 0)[0])/len(y)
    if negetive_ratio > NEGATIVE_RATIO_LIMIT:
        import logging
        import molass_legacy.KekLib.DebugPlot as plt
        logger = logging.getLogger(__name__)
        logger.warning("temporarily correcting a flat negative_baseline due to negetive_ratio=%.3g", negetive_ratio)

        kth = int(len(y)*NEGATIVE_RATIO_LIMIT)
        kpp = np.argpartition(y, kth)
        base = np.average(y[kpp[0:kth]])

        if debug:
            from DataUtils import get_in_folder
            x = np.arange(len(y))
            with plt.Dp():
                fig, ax = plt.subplots()
                ax.set_title("negative baseline fix for %s" % get_in_folder(), fontsize=16)
                ax.plot(x, y, label="data")
                xmin, xmax = ax.get_xlim()
                ax.set_xlim(xmin, xmax)
                ax.plot([xmin, xmax], [0, 0], ":", color="red")
                ax.plot([xmin, xmax], [base, base], color="red")
                fig.tight_layout()
                plt.show()

        y -= base

    return y



"""
    SecTheory.CumsumInverse.py

    Copyright (c) 2022, SAXS Team, KEK-PF
"""
import numpy as np
from scipy.optimize import minimize

PENALTY_SCALE = 1e6

def cumsum_inverse(spline, y_values, xmin, xmax):

    def func(x):
        if len(x) > 1:
            order_penalty = min(0, np.min(x[1:] - x[:-1]))**2
        else:
            order_penalty = 0
        range_penalty = min(0, x[0] - xmin)**2 + min(0, xmax - x[-1])**2
        return np.sum((spline(x) - y_values)**2) + PENALTY_SCALE*(order_penalty + range_penalty)

    n = len(y_values)
    # x_guess = np.linspace(xmin, xmax, n + 2)[1:-1]
    x_guess = xmin*(1 - y_values) + xmax*y_values
    sol = minimize(func, x_guess)
    return sol.x
